from __future__ import print_function

from cloudmesh_client.shell.command import command
from cloudmesh_client.shell.console import Console
from cloudmesh_client.cloud.inventory import Inventory
from cloudmesh_client.common.hostlist import Parameter
from cloudmesh_client.locations import config_file

from cloudmesh_client.shell.command import PluginCommand, CloudPluginCommand


# noinspection PyBroadException
class InventoryCommand(PluginCommand, CloudPluginCommand):
    topics = {"inventory": "todo"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command inventory")

        # TODO: delete row
        # TODO: add columns
        # TODO: ATTRIBUTE=VALUE

    # noinspection PyUnusedLocal
    @command
    def do_inventory(self, args, arguments):
        """
        ::

          Usage:
              inventory add NAMES [--label=LABEL]
                                  [--service=SERVICES]
                                  [--project=PROJECT]
                                  [--owners=OWNERS]
                                  [--comment=COMMENT]
                                  [--cluster=CLUSTER]
                                  [--ip=IP]
              inventory set NAMES for ATTRIBUTE to VALUES
              inventory delete NAMES
              inventory clone NAMES from SOURCE
              inventory list [NAMES] [--format=FORMAT] [--columns=COLUMNS]
              inventory info

          Arguments:

            NAMES     Name of the resources (example i[10-20])

            FORMAT    The format of the output is either txt,
                      yaml, dict, table [default: table].

            OWNERS    a comma separated list of owners for this resource

            LABEL     a unique label for this resource

            SERVICE   a string that identifies the service

            PROJECT   a string that identifies the project

            SOURCE    a single host name to clone from

            COMMENT   a comment
            
          Options:

             -v       verbose mode

          Description:

                add -- adds a resource to the resource inventory

                list -- lists the resources in the given format

                delete -- deletes objects from the table

                clone -- copies the content of an existing object
                         and creates new once with it

                set   -- sets for the specified objects the attribute
                         to the given value or values. If multiple values
                         are used the values are assigned to the and
                         objects in order. See examples

                map   -- allows to set attibutes on a set of objects
                         with a set of values

          Examples:

            cm inventory add x[0-3] --service=openstack

                adds hosts x0, x1, x2, x3 and puts the string
                openstack into the service column

            cm lists

                lists the repository

            cm x[3-4] set temperature to 32

                sets for the resources x3, x4 the value of the
                temperature to 32

            cm x[7-8] set ip 128.0.0.[0-1]

                sets the value of x7 to 128.0.0.0
                sets the value of x8 to 128.0.0.1

            cm clone x[5-6] from x3

                clones the values for x5, x6 from x3

        """
        print(arguments)
        filename = config_file("/cloudmesh_inventory.yaml")

        sorted_keys = True
        if arguments["info"]:
            i = Inventory()
            i.read()
            i.info()
        elif arguments["list"]:
            i = Inventory()
            i.read()
            if arguments["--columns"]:
                order = arguments["--columns"].split(",")
            else:
                order = i.order
            print(i.list(format="table", order=order))
        elif arguments["NAMES"] is None:
            Console.error("Please specify a host name")
        # elif arguments["set"]:
        #    hosts = Parameter.expand_hostlist(arguments["NAMES"])
        #    i = inventory()
        #    i.read()
        #    element = {}

        #    for attribute in i.order:
        #        try:
        #            attribute = arguments["ATTRIBUTE"]
        #            value = arguments["VALUE"]
        #            if value is not None:
        #                element[attribute] = value
        #        except:
        #            pass
        #    element['host'] = arguments["NAMES"]
        #    i.add(**element)
        #    print (i.list(format="table"))
        elif arguments["set"]:
            hosts = Parameter.expand(arguments["NAMES"])
            values = Parameter.expand(arguments["VALUES"])
            if len(values) == 1:
                values *= len(hosts)
            print(hosts)
            print(values)
            attribute = arguments["ATTRIBUTE"]
            if len(hosts) != len(values):
                Console.error(
                    "Number of names {:} != number of values{:}".format(
                        len(hosts), len(values)))

            i = Inventory()
            i.read()

            for index in range(0, len(hosts)):
                host = hosts[index]
                value = values[index]

                host_object = {'host': host, attribute: value}

                i.add(**host_object)

            print(i.list(format="table"))
        elif arguments["add"]:
            hosts = Parameter.expand(arguments["NAMES"])
            i = Inventory()
            i.read()
            element = {}

            for attribute in i.order:
                try:
                    value = arguments["--" + attribute]
                    if value is not None:
                        element[attribute] = value
                except:
                    pass
            element['host'] = arguments["NAMES"]
            i.add(**element)
            print(i.list(format="table"))
        elif arguments["delete"]:
            hosts = Parameter.expand(arguments["NAMES"])
            i = Inventory()
            i.read()

            for host in hosts:
                del i.data[host]
            i.save()
        elif arguments["clone"]:
            hosts = Parameter.expand(arguments["NAMES"])
            source = arguments["SOURCE"]

            i = Inventory()
            i.read()

            if source in i.data:

                for host in hosts:
                    i.data[host] = dict(i.data[source])
                i.save()
            else:
                Console.error("The source {:} does not exist".format(source))
        return ""




import importlib

class DynamicClass(object):

    def add_instance_method(self, f):
        setattr(self, f.__name__, f)

    @classmethod
    def add_classmethod(cls, f):
        setattr(cls, f.__name__, f)

    def load_instancemethod(self, location):
        module_name, class_name = location.rsplit(".", 1)
        f = getattr(importlib.import_module(module_name), class_name)
        self.add_instance_method(f)

    @classmethod
    def load_classmethod(cls, location):
        module_name, class_name = location.rsplit(".", 1)
        f = getattr(importlib.import_module(module_name), class_name)
        cls.add_classmethod(f)


from __future__ import print_function
from cloudmesh_client.shell.command import command, PluginCommand, CloudPluginCommand
from cloudmesh_client.shell.console import Console
from cloudmesh_client.default import Default


class CheckCommand(PluginCommand, CloudPluginCommand):
    topics = {"check": "notimplemented"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command check")

    # noinspection PyUnusedLocal
    @command
    def do_check(self, args, arguments):
        """
        ::

            Usage:
                check --cloud=CLOUD
                check

                checks some elementary setting for cloudmesh

            Options:
               --format=FORMAT  the output format [default: table]
               --cloud=CLOUD    the cloud name

            Examples:
                cm check
                cm check --cloud=kilo

        """
        cloud = arguments["--cloud"] or Default.cloud

        if cloud is None:
            Console.error("Default cloud doesn't exist")

        print(locals())

        Console.TODO("This command is not implemented yet", traceflag=False)
        Console.ok("{:} ok".format(cloud))

        return ""



from __future__ import print_function
# from cloudmesh_client.shell.console import Console

from cloudmesh_client.db import CloudmeshDatabase

from cloudmesh_client.common.todo import TODO

from cloudmesh_client.cloud.ListResource import ListResource


class Reservation(ListResource):
    cm = CloudmeshDatabase()

    def info(cls, user=None, project=None):
        """
        prints if the user has access to the reservation an on which host.

        :param user:
        :param project:
        :return:
        """
        TODO.implement()

    def add_from_file(cls, filename):
        """

        :param filename:
        :return:
        """
        TODO.implement()

    @classmethod
    def add(cls,
            name,
            start,
            end,
            user=None,
            project=None,
            hosts=None,
            description=None,
            category=None):
        """

        :param name: Name of reservation
        :param start: Start time of reservation
        :param end: End time of reservation
        :param user: Reserved by this user
        :param project: Reservation project
        :param hosts: Reserved hosts
        :param description: Description
        :param cloud: Cloud into which reservation done
        :return:
        """

        o = {
            "provider": "general",
            "kind": "reservation",
            "name": name,
            "hosts": hosts,
            "start": start,
            "end": end,
            "description": description,
            "category": category,
            "user": user,
            "project": project
        }

        cls.cm.add(o)

    @classmethod
    def delete(cls,
               name=None,
               start=None,
               end=None,
               user=None,
               project=None,
               hosts=None):
        """

        :param name: Name of reservation
        :param start: Start time of reservation
        :param end: End time of reservation
        :param user: Reserved by this user
        :param project: Reservation project
        :param hosts: Hosts reserved
        :return:
        """

        args = {}

        if name is not None:
            args['name'] = name
        if start is not None:
            args['start_time'] = start
        if end is not None:
            args['end_time'] = end
        if user is not None:
            args['user'] = user
        if project is not None:
            args['project'] = project
        if hosts is not None:
            args['hosts'] = hosts

        # TODO: Improve this logic
        result = cls.cm.find(kind="reservation", provider="general", output="dict", scope='all', **args)
        if result is not None:
            for res in result:
                cls.cm.delete(kind="reservation", provider="general", name=res["name"])

    @classmethod
    def delete_from_file(cls, filename):
        """

        :param filename:
        :return:
        """
        TODO.implement()

    @classmethod
    def suspend(cls, names=None):
        TODO.implement()

    @classmethod
    def clear(cls, names=None):
        TODO.implement()

    @classmethod
    def refresh(cls, names=None):
        TODO.implement()

    @classmethod
    def resume(cls, names=None):
        TODO.implement()

    @classmethod
    def list(cls,
             name=None,
             start=None,
             end=None,
             user=None,
             project=None,
             hosts=None):

        """

        :param name: Name of reservation
        :param start: Start time of reservation
        :param end: End time of reservation
        :param user: Reserved by this user
        :param project: Reservation project
        :param hosts: Hosts reserved
        :return:
        """

        args = {}

        if name is not None:
            args['name'] = name
        if start is not None:
            args['start_time'] = start
        if end is not None:
            args['end_time'] = end
        if user is not None:
            args['user'] = user
        if project is not None:
            args['project'] = project
        if hosts is not None:
            args['hosts'] = hosts
        args["kind"] = "reservation"

        # print(args)
        result = cls.cm.find(**args)
        # print("RESULT:- {}".format(result))
        return result

    @classmethod
    def update(cls,
               name,
               start,
               end,
               user=None,
               project=None,
               hosts=None,
               description=None,
               cloud=None):

        entry = {
            'start_time': start,
            'end_time': end,
            'name': name,
            'user': user,
            'project': project,
            'hosts': hosts,
            'description': description,
            'cloud': cloud,
            "kind": "reservation",
            "provider": "general"
        }

        update = dict(entry)
        for key in entry:
            if entry[key] is None:
                del update[key]

        print ("ARGS", update)
        cls.cm.add(entry)



class Attributes(object):
    @classmethod
    def get(cls, kind):
        layout = {
            'flavor': {
                'order': [
                    'id',
                    'name',
                    'user',
                    'ram',
                    'os_flv_disabled',
                    'vcpus',
                    'swap',
                    'os_flavor_acces',
                    'rxtx_factor',
                    'os_flv_ext_data',
                    'disk',
                    'category',
                    'uuid',
                    'updated_at'
                ],
                'header': [
                    'Id',
                    'Name',
                    'User',
                    'RAM',
                    'Disabled',
                    'vCPUs',
                    'Swap',
                    'Access',
                    'rxtx_factor',
                    'os_flv_ext_data',
                    'Disk',
                    'Cloud',
                    'UUID',
                    'updated'
                ]
            },
            'image': {
                'order': [
                    'id',
                    'name',
                    'os_image_size',
                    'metadata__description',
                    'minDisk',
                    'minRam',
                    'progress',
                    'status',
                    'updated',
                    'uuid',
                    'category',
                    'updated_at'
                ],
                'header': [
                    'id',
                    'name',
                    'size',
                    'description',
                    'minDisk',
                    'minRam',
                    'progress',
                    'status',
                    'updated',
                    'uuid',
                    'cloud',
                    'updated_at'
                ]
            },
            'vm': {
                'order': [
                    'id',
                    'group',
                    'label',
                    'status',
                    'static_ip',
                    'floating_ip',
                    'key',
                    'project',
                    'user',
                    'category',
                    'updated_at'
                ],
                'header': [
                    'id',
                    'group',
                    'label',
                    'status',
                    'static_ip',
                    'floating_ip',
                    'key',
                    'project',
                    'user',
                    'cloud',
                    'updated_at'
                ]
            },
            'floating_ip': {
                'order': [
                    "instance_name",
                    "ip",
                    "pool",
                    "fixed_ip",
                    "id",
                    "instance_id",
                    'cloud'
                ],
                'header': [
                    "instance_name",
                    "floating_ip",
                    "floating_ip_pool",
                    "fixed_ip",
                    "floating_ip_id",
                    "instance_id",
                    'cloud'
                ],
            },
            'floating_ip_pool': {
                'order': [
                    "name"
                ],
                'header': [
                    "floating_ip_pool"
                ],
            },
            'clouds': {
                'order': [
                    "id",
                    "cloud",
                    "default",
                    "active",
                    "status",
                    "key"
                ],
                'header': [
                    "id",
                    "Cloud",
                    "Default",
                    "Active",
                    "Status",
                    "Key"
                ],
            },
            'limits': {
                'order': [
                    'Name',
                    'Value'
                ],
                'header': [
                    'Name',
                    'Value'
                ]
            },
            'quota': {
                'order': [
                    'Quota',
                    'Limit'
                ],
                'header': [
                    'Quota',
                    'Limit'
                ]
            },
            'secgroup': {
                'order': [
                    'id',
                    'name',
                    'category',
                    'user',
                    'project',
                    'uuid'
                ],
                'header': [
                    'id',
                    'secgroup_name',
                    'category',
                    'user',
                    'tenant_id',
                    'secgroup_uuid'
                ]
            },
            'default': {
                'order': [
                    'user',
                    'category',
                    'name',
                    'value',
                    'created_at',
                    'updated_at'
                ],
                'header': [
                    'user',
                    'category',
                    'name',
                    'value',
                    'created_at',
                    'updated_at'
                ],
            },
            'group': {
                'order': [
                    "name",
                    "member",
                    "user",
                    "category",
                    "type"],
                'header': [
                    "name",
                    "member",
                    "user",
                    "category",
                    "type"]

            },
            'key': {
                'order': [
                    'keypair__name',
                    "type",
                    "comment",
                    "keypair__fingerprint"
                ],
                'header': [
                    "Name",
                    "Type",
                    "Comment",
                    "Fingerprint"]
            }
        }

        if kind in layout:
            order = layout[kind]["order"]
            header = layout[kind]["header"]
        else:
            order = None
            header = None

        return order, header



from __future__ import print_function
from cloudmesh_client.shell.command import command, PluginCommand, CloudPluginCommand
from cloudmesh_client.shell.console import Console
from cloudmesh_client.default import Default


class TimerCommand(PluginCommand, CloudPluginCommand):
    topics = {"timer": "shell"}

    def __init__(self, context):
        self.context = context
        try:
            if self.context.timer:
                print("init command timer")
        except:
            self.context.timer = False
            # try:
            #    value = Default.get_timer()
            # except:
            #    Default.set_timer("off")

    @command
    def do_timer(self, args, arguments):
        """
        ::

            Usage:
                timer on
                timer off
                timer list [NAME]
                timer start NAME
                timer stop NAME
                timer resume NAME
                timer reset [NAME]

            Description:

                 timer on | off
                     switches timers on and off not yet implemented.
                     If the timer is on each command will be timed and its
                     time is printed after the command. Please note that
                     background command times are not added.

                timer list
                    list all timers

                timer start NAME
                    starts the timer with the name. A start resets the timer to 0.

                timer stop NAME
                    stops the timer

                timer resume NAME
                    resumes the timer

                timer reset NAME
                    resets the named timer to 0. If no name is specified all
                    timers are reset

        """
        # print arguments
        # print "args", args

        if arguments["on"]:
            Default.set_timer("on")
            Console.ok("Switch timer on")
        elif arguments["off"]:
            Default.set_timer("off")
            Console.ok("Switch timer off")
        elif arguments["list"]:

            name = arguments("NAME")
            if name is None:
                print(self.watch)
            else:
                value = self.watch.get(name)
                Console.ok("Timer ({}): {}".format(name, value))
        elif arguments["start"]:
            name = arguments("NAME")
            self.watch.start(name)
            Console.ok("Start timer", name)
        elif arguments["stop"]:
            name = arguments("NAME")
            self.watch.stop(name)
            Console.ok("Stop timer", name)
        elif arguments["reset"]:
            name = arguments("NAME")
            self.watch.reset(name)
            Console.ok("Reset timer", name)



""" run with

python setup.py install; nosetests -v --nocapture  tests/cm_libcloud/test_libcloud_api.py:Test_libcloud_api.test_001

nosetests -v --nocapture tests/libcloud/test_libcloud_api.py

or

nosetests -v tests/test_image.py

"""

from cloudmesh_client.cloud.iaas.CloudProvider import CloudProvider
from cloudmesh_client.common.util import HEADING


# noinspection PyMethodMayBeStatic,PyPep8Naming
class Test_libcloud_api(object):
    """
        This class tests the ImageCommand
    """

    '''
    def test_001(self):


        from libcloud.compute.types import Provider
        from libcloud.compute.providers import get_driver

        config = ConfigDict('cloudmesh.yaml')
        # print (config)
        credential = dict(config["cloudmesh"]["clouds"]["chameleon-ec2"]["credentials"])
        pprint (dict(credential))
        EC2_ACCESS_KEY = credential['EC2_ACCESS_KEY']
        EC2_SECRET_KEY = credential['EC2_SECRET_KEY']

        Driver = get_driver(Provider.EC2)
        conn = Driver(EC2_ACCESS_KEY, EC2_SECRET_KEY)

        keys = conn.list_images()

        print (keys)


    '''

    def test_001(self):
        """test image list :return: """
        HEADING()

        from pprint import pprint

        cloud = "chameleon-ec2"
        provider = CloudProvider(cloud).provider

        print(provider, type(provider))

        # pprint (provider.__dict__)
        # pprint (dir(provider))

        # r = provider.list_flavor(cloud)
        # pprint(r)

        for kind in ["image"]:  # ["image", "vm", "flavor", "key"]: # , "flavor", "vm", "limits", "quota"]:
            r = provider.list(kind, cloud)
            pprint(r)

        assert True

    def test_002(self):
        """ test flavor list :return:  """
        HEADING()

        from pprint import pprint

        cloud = "chameleon-ec2"
        provider = CloudProvider(cloud).provider

        # print (provider, type(provider))

        # r = provider.list_flavor(cloud)
        # pprint(r)

        kind = 'flavor'

        r = provider.list(kind, cloud)
        pprint(r)

        assert 't2.small' in str(r)

        r = provider.list_flavor(cloud)
        pprint(r)

        assert 't2.small' in str(r)

        r = provider.provider.list_sizes(cloud)
        pprint(r)

        assert 't2.small' in str(r)

    def test_003(self):
        """ test vm list:return:  """
        HEADING()

        from pprint import pprint

        cloud = "chameleon-ec2"
        provider = CloudProvider(cloud).provider

        print(provider, type(provider))

        # pprint (provider.__dict__)
        # pprint (dir(provider))

        # r = provider.list_flavor(cloud)
        # pprint(r)

        for kind in ["vm"]:  # ["image", "vm", "flavor", "key"]: # , "flavor", "vm", "limits", "quota"]:
            r = provider.list(kind, cloud)
            pprint(r)

        assert True

    def test_004(self):
        """ test key list:return:"""
        HEADING()

        from pprint import pprint

        cloud = "chameleon-ec2"
        provider = CloudProvider(cloud).provider

        print(provider, type(provider))

        # pprint (provider.__dict__)
        # pprint (dir(provider))

        # r = provider.list_flavor(cloud)
        # pprint(r)

        for kind in ["key"]:  # ["image", "vm", "flavor", "key"]: # , "flavor", "vm", "limits", "quota"]:
            r = provider.list(kind, cloud)
            pprint(r)

        assert True




from cloudmesh_client.shell.command import PluginCommand, CloudPluginCommand
from cloudmesh_client.common.dotdict import dotdict
from pprint import pprint
from cloudmesh_client.shell.command import command
from cloudmesh_client.cloud.key import Key
from cloudmesh_client.shell.console import Console

class AkeyCommand(PluginCommand, CloudPluginCommand):
    topics = {"akey": "security"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command akey")

    @command
    def do_akey(self, args, arguments):
        """
        ::

           Usage:
           akey
           akey list
           akey add --name=key-name --pub=pub-key-path --cert=certificate-file-path --pfx=pfx-file-path
        """
        # pprint("Akey command pressed")
        # pprint(args)
        # pprint(arguments)
        if arguments['list']:
            print("Key list time")
        elif arguments['add']:
            Console.info("Azure key addition invoked")
            if arguments['--name']:
                print("name:"+arguments['--name'])
                key_name = arguments['--name']
            if arguments['--cert']:
                print("cert:"+arguments['--cert'])
                certificate_path = arguments['--cert']
            if arguments['--pub']:
                print("pub:"+arguments['--pub'])
                key_path = arguments['--pub']
            if arguments['--pfx']:
                print("pfx:"+arguments['--pfx'])
                pfx_path = arguments['--pfx']
            Key.add_azure_key_to_db(key_name, key_path, certificate_path, pfx_path)
        return ""



""" run with

python setup.py install; nosetests -v --nocapture  tests/cm_cloud/test_limits.py:Test_limits.test_001

nosetests -v --nocapture tests/test_limits.py

or

nosetests -v tests/test_limits.py

"""

from cloudmesh_client.common.ConfigDict import ConfigDict
from cloudmesh_client.common.Shell import Shell
from cloudmesh_client.common.dotdict import dotdict
from cloudmesh_client.common.util import HEADING
from cloudmesh_client.common.util import banner
from cloudmesh_client.default import Default


# noinspection PyPep8Naming
class Test_limits:
    """
        This class tests the LimitsCommand
    """

    data = dotdict({
        "cloud": Default.cloud,
        "wrong_cloud": "no_cloud",
        "tenant": "TBD",
        "format": 'table'
    })
    config = ConfigDict("cloudmesh.yaml")
    credentials = config["cloudmesh"]["clouds"][data.cloud]["credentials"]
    data.tenant = credentials.get("OS_TENANT_NAME") or credentials.get("OS_TENANT_ID")

    def run(self, command):
        command = command.format(**self.data)
        banner(command, c="-")
        print(command)
        parameter = command.split(" ")
        shell_command = parameter[0]
        args = parameter[1:]
        result = Shell.execute(shell_command, args)
        print(result)
        return result

    def setup(self):
        pass

    # noinspection PyPep8Naming
    def tearDown(self):
        pass

    def test_001(self):
        HEADING("test limits list")
        result = self.run("cm limits list --cloud={cloud}")
        assert "Name" in result

    def test_002(self):
        HEADING("test limits list with csv output")
        result = self.run("cm limits list --format={format}").split('\n')
        assert "maxTotalFloatingIps" in str(result)

    def test_003(self):
        HEADING("test limits class where cloud doesnt exist")
        result = self.run("cm limits list --cloud={wrong_cloud}")
        assert "ERROR" in str(result)

#   --tenant=
#   not allowed on chameleon cloud, so we outcomment.
#
#    def test_004(self):
#        HEADING("test limits class with unauthorised access")
#        result = self.run("cm limits list --tenant={tenant}")
#        assert "Not authorized" in str(result)



from __future__ import print_function
from cloudmesh_client.shell.command import command


class BarCommand(object):
    topics = {"bar": "example"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init Bar")

    # noinspection PyUnusedLocal
    @command
    def do_bar(self, args, arguments):
        """
        ::

          Usage:
                bar -f FILE
                bar FILE
                bar list

          This command does some useful things.

          Arguments:
              FILE   a file name

          Options:
              -f      specify the file

        """
        print(arguments)



from __future__ import print_function

import getpass
import hashlib
import json
import os
import random
import signal
import string
import sys
import time
import webbrowser
from builtins import input
from pprint import pprint


import requests
from requests.auth import HTTPBasicAuth
from httpsig.requests_auth import HTTPSignatureAuth
from requests.auth import HTTPBasicAuth

from cloudmesh_client.common.ConfigDict import ConfigDict
from cloudmesh_client.common.Shell import Shell
from cloudmesh_client.common.util import banner
from cloudmesh_client.shell.console import Console

requests.packages.urllib3.disable_warnings()


# noinspection PyBroadException,PyBroadException
class Comet(object):
    # api_version = "/v1"
    endpoint = ''
    base_uri = ''
    api_version = ''
    # base_uri = "https://comet-nucleus.sdsc.edu/nucleus"
    local_base_uri = "https://localhost:8443/nucleus"
    auth_uri = "{}/rest-auth".format(base_uri)
    local_auth_uri = "{}/rest-auth".format(local_base_uri)
    tunnelled = False
    # "USERPASS", "APIKEY"
    auth_provider = None
    token = None
    api_key = None
    api_secret = None
    api_auth = None
    HEADER = {'content-type': 'application/json'}
    AUTH_HEADER = {'content-type': 'application/json'}

    # in case of https endpoint
    verify = False

    @staticmethod
    def set_endpoint(endpoint):
        Comet.endpoint = endpoint

    @staticmethod
    def set_base_uri(uri):
        Comet.base_uri = uri
        Comet.auth_uri = Comet.base_uri + "/rest-auth"

    @staticmethod
    def set_api_version(api_version):
        Comet.api_version = "/%s" % api_version

    @staticmethod
    def url(path):
        if Comet.tunnelled:
            url = Comet.local_base_uri + Comet.api_version + "/" + path
        else:
            url = Comet.base_uri + Comet.api_version + "/" + path
        return url

    def __init__(self):
        #
        # TODO: in future set global uris
        #
        pass

    @staticmethod
    def docs():

        webbrowser.open("{}/docs/#!/v1".format(Comet.base_uri))

    # #####################
    # TUNNEL
    # #####################

    @staticmethod
    def tunnel(start):
        if start:
            Comet.tunnelled = True
            command = "ssh -L 8443:localhost:443 nucleus"
            os.system(command)
        else:
            Comet.kill_tunnel()

    @staticmethod
    def kill_tunnel():
        pid = Comet.find_tunnel()
        if pid is None:
            Console.error("No tunnel to comet found")
        else:
            Console.ok("Killing the tunnel to comet")
            os.kill(pid, signal.SIGTERM)

    @staticmethod
    def state():
        pid = Comet.find_tunnel()
        Console.ok("Comet tunnel: {:}".format(pid))

    @staticmethod
    def is_tunnel():
        pid = Comet.find_tunnel()
        return pid is not None

    @staticmethod
    def find_tunnel():
        r = Shell.execute("ps", ["-ax"]).split("\n")
        pid = None
        info = None
        for line in r:
            if ("localhost" in line and "nucleus" in line) or (
                            "comet" in line and "tunnel" in line) and 'status' not in line:
                info = line.strip()
                break
        if info:
            pid = int(info.split(" ", 1)[0])

        return pid

    # ##############################
    # AUTHENTICATE
    # ##############################
    '''
    import json
    import requests
    from httpsig.requests_auth import HTTPSignatureAuth
    import time, random, string

    secret = "REALAPISECRET"

    auth = HTTPSignatureAuth(secret=secret, headers=["nonce", "timestamp"])

    nonce = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(10))

    nodes2op = {"computes":"vm-vc2-[6-7]","cluster":"vc2"}
    z = requests.post('https://comet-nucleus.sdsc.edu/nucleus/v1/computeset/',
        data=json.dumps(nodes2op), auth=auth, headers={"timestamp": int(time.time()),
        "nonce":nonce, "X-Api-Key":"Fugang"})
    print z.status_code
    print z.headers
    print z.text

    '''

    @staticmethod
    def get_nonce():
        nonce = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(10))
        return nonce

    @classmethod
    def set_auth_provider(cls, auth_provider=None):
        # try to load from yaml file if not specified
        if not auth_provider:
            config = ConfigDict("cloudmesh.yaml")
            cometConf = config["cloudmesh.comet"]
            auth_provider = cometConf["endpoints"][cls.endpoint]["auth_provider"].upper()
            # value not set in yaml file, use USERPASS as default
            if not auth_provider:
                auth_provider = "USERPASS"
        cls.auth_provider = auth_provider

    @classmethod
    def logon(cls, endpoint=None, username=None, password=None):
        config = ConfigDict("cloudmesh.yaml")
        cometConf = config["cloudmesh.comet"]
        if endpoint:
            cls.set_endpoint(endpoint)
        else:
            cls.set_endpoint(cometConf["active"])
        cls.set_base_uri(cometConf["endpoints"][cls.endpoint]["nucleus_base_url"])
        cls.set_api_version(cometConf["endpoints"][cls.endpoint]["api_version"])
        if not cls.auth_provider:
            cls.set_auth_provider()
        # print (cls.endpoint)
        # print (cls.base_uri)
        # print (cls.api_version)
        # print (cls.auth_provider)
        ret = False
        if "USERPASS" == cls.auth_provider:
            # for unit testing only.
            if username is None:
                username = cometConf["endpoints"][cls.endpoint]["userpass"]["username"]
                if username == '' or username == 'TBD':
                    username = cometConf["username"]
            if password is None:
                password = cometConf["endpoints"][cls.endpoint]["userpass"]["password"]
                if password.lower() == "readline":
                    password = getpass.getpass()
                elif password.lower() == "env":
                    password = os.environ.get("COMET_PASSWORD", getpass.getpass())

            if cls.token is None:
                if cls.auth_uri:
                    if cls.tunnelled:
                        authuri = "%s/login/" % cls.local_auth_uri
                    else:
                        authuri = "%s/login/" % cls.auth_uri
                    data = {"username": username, "password": password}
                    r = requests.post(authuri,
                                      data=json.dumps(data),
                                      headers=cls.HEADER,
                                      verify=cls.verify)
                    try:
                        cls.token = r.json()["key"]
                        cls.AUTH_HEADER['Authorization'] = "Token {:}".format(
                            cls.token)
                    except:
                        ret = False
                    ret = cls.token
            else:
                ret = cls.token
        elif "APIKEY" == cls.auth_provider:
            # print ("API KEY based auth goes here")
            cls.api_key = cometConf["endpoints"][cls.endpoint]["apikey"]["api_key"]
            cls.api_secret = cometConf["endpoints"][cls.endpoint]["apikey"]["api_secret"]
            cls.api_auth = HTTPSignatureAuth(secret=cls.api_secret, headers=["nonce", "timestamp"])
            #
            # api key based auth does not maintain a session
            # once values specified, considered as AuthNed.
            if cls.api_key and cls.api_secret and cls.api_auth:
                ret = True
        else:
            print("The specified AUTH Provider Not Currently Supported")
            pass
        return ret

    @classmethod
    def logoff(cls):
        ret = True
        if "USERPASS" == cls.auth_provider:
            if cls.token:
                if cls.auth_uri:
                    authuri = "%s/logout/" % cls.auth_uri
                    header = dict(cls.HEADER)
                    header['Authorization'] = "Token %s" % cls.token
                    r = requests.post(authuri,
                                      headers=header,
                                      verify=cls.verify)
                    cls.token = None
                    cls.AUTH_HEADER = cls.HEADER
                else:
                    ret = False
        return ret

    @classmethod
    def status(cls):
        ret = True
        if "USERPASS" == cls.auth_provider and cls.token is None:
            ret = False
        return ret

    # #############################
    # GET FROM REST
    # #############################

    # To make GET calls for synchronous or asynchronous API

    @staticmethod
    def get(url, headers=None, allow_redirects=True, data=None,
                 authuser=None, authpass=None):
        return Comet.http(url, action="get", headers=headers, data=data,
                          authuser=authuser, authpass=authpass,
                          allow_redirects=allow_redirects)

    @staticmethod
    def post(url, headers=None, data=None, md5=None, files=None,
             cacert=True, allow_redirects=True):
        return Comet.http(url, action="post", headers=headers, data=data,
                          files=files, md5=md5, cacert=cacert,
                          allow_redirects=allow_redirects)

    @staticmethod
    def put(url, headers=None, data=None, allow_redirects=True):
        return Comet.http(url, action="put", headers=headers,
                          data=data, allow_redirects=allow_redirects)

    # To make GET calls for synchronous or asynchronous API
    @staticmethod
    def http(url, action="get",
             headers=None, data=None, authuser=None, authpass=None,
             files=None, md5=None, cacert=True, allow_redirects=True):
        # print ("KKK", url)
        # print ("KKK", action)
        # print ("KKK", Comet.auth_provider)
        ret = None
        if Comet.tunnelled:
            cacert = False
        if "USERPASS" == Comet.auth_provider:
            if headers is None:
                headers = Comet.AUTH_HEADER
            if 'post' == action:
                if files:
                    del headers["content-type"]
                    headers["md5"] = md5
                    r = requests.post(url, headers=headers, files=files,
                                      allow_redirects=allow_redirects,
                                      verify=cacert)
                else:
                    r = requests.post(url, headers=headers, data=json.dumps(data),
                                      allow_redirects=allow_redirects,
                                      verify=cacert)

            elif 'put' == action:
                r = requests.put(url, headers=headers, data=json.dumps(data),
                                 allow_redirects=allow_redirects, verify=cacert)
            else:
                if data:
                    if authuser and authpass:
                        r = requests.get(url, headers=headers, params=data,
                                         auth=(authuser, authpass),
                                         allow_redirects=allow_redirects, verify=cacert)
                    else:
                        r = requests.get(url, headers=headers, params=data,
                                         allow_redirects=allow_redirects, verify=cacert)
                else:
                    if authuser and authpass:
                        r = requests.get(url, headers=headers,
                                         auth=(authuser, authpass),
                                         allow_redirects=allow_redirects, verify=cacert)
                    else:
                        r = requests.get(url, headers=headers,
                                         allow_redirects=allow_redirects, verify=cacert)

            #print ("KKK --- DEBUGGING HTTP CALL (USERPASS)")
            #pprint (r)
            #pprint (r.status_code)
            #pprint (r.headers)
            #pprint (r.text)

            # 303 redirect
            if r.status_code == 303:
                ret = r.headers["Location"]
            # responded immediately
            elif r.status_code == 200:
                try:
                    ret = r.json()
                except:
                    ret = r.text
            # processed successfully, but returned empty
            elif r.status_code == 204:
                ret = ''
            # code 201, created for the post request
            elif r.status_code == 201:
                # now automatically redirect to result page
                # thus no need to check status periodically.
                # Currently it works well for cluster listing
                # However not sure if the delay is large, what the behaviour
                # would be
                finished = False
                newurl = r.headers["Location"]
                while not finished:
                    ret = requests.get(newurl, headers=headers)
                    try:
                        ret = ret.json()
                    except:
                        pass
                    # in some occasions, when the result is not ready,
                    # the result still has 'status' in it (value as '0')
                    # otherwise it's the correct value after redirection
                    if 'status' not in ret:
                        finished = True
                    else:
                        time.sleep(1)
            elif r.status_code == 401:
                ret = {"error": "Not Authenticated"}
            elif r.status_code == 403:
                ret = {"error": "Permission denied"}
            elif r.status_code == 400:
                ret = {"error": "%s" % r.text}
        elif "APIKEY" == Comet.auth_provider:
            headers = {'content-type': 'application/json',
                       "timestamp": int(time.time()),
                       "nonce": Comet.get_nonce(),
                       "X-Api-Key": Comet.api_key}
            # print ("KKK", headers)
            # print ("KKK", Comet.api_auth)
            # print ("KKK", data)
            # print ("KKK", cacert)
            if 'post' == action:
                if files:
                    headers = {"timestamp": int(time.time()),
                               "nonce": Comet.get_nonce(),
                               "X-Api-Key": Comet.api_key,
                               "md5": md5}
                    r = requests.post(url, auth=Comet.api_auth, headers=headers,
                                      files=files,
                                      allow_redirects=allow_redirects,
                                      verify=cacert)
                else:
                    r = requests.post(url, auth=Comet.api_auth, headers=headers,
                                      data=json.dumps(data),
                                      allow_redirects=allow_redirects,
                                      verify=cacert)
            elif 'put' == action:
                r = requests.put(url, auth=Comet.api_auth, headers=headers,
                                 data=json.dumps(data),
                                 allow_redirects=allow_redirects,
                                 verify=cacert)
            else:
                if data:
                    # print (url)
                    # print (data)
                    r = requests.get(url, auth=Comet.api_auth, headers=headers,
                                     params=data,
                                     allow_redirects=allow_redirects,
                                     verify=cacert)
                else:
                    r = requests.get(url, auth=Comet.api_auth, headers=headers,
                                     allow_redirects=allow_redirects,
                                     verify=cacert)
            # pprint (r.url)
            ret = None

            #print ("KKK --- DEBUGGING HTTP CALL (APIKEY)")
            #pprint (r)
            #pprint (r.status_code)
            #pprint (r.headers)
            #pprint (r.text)

            # 303 redirect
            if r.status_code == 303:
                ret = r.headers["Location"]
            # responded immediately
            elif r.status_code == 200:
                try:
                    ret = r.json()
                except:
                    ret = r.text
            # processed successfully, but returned empty
            elif r.status_code == 204:
                ret = ''
            # code 201, created for the post request
            elif r.status_code == 201:
                finished = False
                newurl = r.headers["Location"]
                headers["timestamp"] = int(time.time())
                headers["nonce"] = Comet.get_nonce()
                while not finished:
                    ret = requests.get(newurl, auth=Comet.api_auth, headers=headers)
                    try:
                        ret = ret.json()
                    except:
                        pass
                    if 'status' not in ret:
                        finished = True
                    else:
                        time.sleep(1)
            elif r.status_code == 401:
                try:
                    ret = r.json()
                    ret = {"error": ret}
                except:
                    pass
                if not ret:
                    ret = {"error": "Not Authenticated"}
            elif r.status_code == 403:
                try:
                    ret = r.json()
                    ret = {"error": ret}
                except:
                    pass
                if not ret:
                    ret = {"error": "Permission denied"}
            elif r.status_code == 400:
                try:
                    ret = r.json()
                    ret = {"error": ret}
                except:
                    pass
                if not ret:
                    ret = {"error": "%s" % r.text}
        # print ("KKKKK", ret)
        return ret

    @staticmethod
    def get_computeset(id=None, state=None):
        # print (id, state)
        if not id:
            if not state:
                state = 'running'
            params = {'state': state}
            geturl = Comet.url("computeset/")
            r = Comet.get(geturl, data=params)
            # pprint (r)
        else:
            geturl = Comet.url("computeset/{}/".format(id))
            r = Comet.get(geturl)
        return r

    @staticmethod
    def console_url(clusterid, nodeid=None):
        config = ConfigDict("cloudmesh.yaml")
        cometConf = config["cloudmesh.comet"]
        defaultUser = cometConf["username"]
        '''
        user = input("Comet nucleus username [%s]: " \
                     % defaultUser)
        if not user:
            user = defaultUser
        '''
        user = None
        if defaultUser and 'TBD'!=defaultUser:
            user = defaultUser
        else:
            user = input("Enter comet nucleus username: ")
        password = getpass.getpass("Enter comet nucleus password: ")
        return_url = None
        # console access requires 2-factor, and only supported by userpass
        Comet.set_auth_provider(auth_provider="USERPASS")
        Comet.logon()
        # print (Comet.auth_provider)
        if not nodeid:
            url = Comet.url("cluster/{}/frontend/console/".format(clusterid))
        else:
            url = Comet.url("cluster/{}/compute/{}/console/".format(clusterid, nodeid))
        return_url = Comet.get(url, authuser=user, authpass=password, allow_redirects=False)

        # restore the preset auth method
        auth_provider = cometConf["endpoints"][Comet.endpoint]["auth_provider"].upper()
        Comet.set_auth_provider(auth_provider=auth_provider)
        Comet.logon()
        # print (Comet.auth_provider)
        # print ("KKK", return_url)
        '''
        r = None
        if "USERPASS" == Comet.auth_provider:
            r = requests.get(url,
                             headers=Comet.AUTH_HEADER,
                             allow_redirects=False,
                             verify=True)
        elif "APIKEY" == Comet.auth_provider:
            headers = {'content-type': 'application/json',
                       "timestamp": int(time.time()),
                       "nonce": Comet.get_nonce(),
                       "X-Api-Key": Comet.api_key}
            r = requests.get(url,
                             auth=Comet.api_auth,
                             headers=headers,
                             allow_redirects=False,
                             verify=True)
        # print (url)
        # print (r.status_code)
        # print (r.headers)
        # print (r.text)
        if r and 303 == r.status_code:
            return_url = r.headers["Location"]
        '''
        return return_url

    @staticmethod
    def console(clusterid, nodeid=None, linkonly=False):
        url = Comet.console_url(clusterid, nodeid)
        #pprint (url)
        if url:
            if 'error' in url:
                Console.error(url['error'], traceflag=False)
            else:
                newurl_esc = url.replace("&", "\&")
                print ("Console URL: {}".format(url))
                if not linkonly:
                    # for OSX
                    if 'darwin' == sys.platform:
                        os.system("open {}".format(newurl_esc))
                    # for linux - tested on Ubuntu 14.04 and CentOS 7.1
                    elif 'linux2' == sys.platform:
                        os.system("firefox {} &".format(newurl_esc))
                    else:
                        Console.error("No supportted OS/browser detected!"
                                      "Use the above url manually in your brower:\n",
                                      traceflag=False)
        else:
            Console.error("Console URL not available."\
                          "Please make sure the node is running and try again!",
                          traceflag=False)

    @staticmethod
    def md5(fname):
        hash = hashlib.md5()
        with open(fname, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash.update(chunk)
        return hash.hexdigest()

    @staticmethod
    def list_iso():
        ret = ''
        url = Comet.url("image")
        r = Comet.get(url)
        if r is not None:
            ret = r
        return ret

    @staticmethod
    def upload_iso(filename, filepath):
        ret = ''
        # print ("filename to use: %s" % filename)
        # print ("full file path: %s" % filepath)
        posturl = Comet.url("image")
        r = None
        md5 = Comet.md5(filepath)
        with open(filepath, 'rb') as fh:
            files = {'file': (filename, fh)}
            print("File to be uploaded: %s" % filename)
            r = Comet.post(posturl, files=files, md5=md5)
            if r is not None:
                ret = r
        return ret

    @staticmethod
    def get_apikey(endpoint):
        config = ConfigDict("cloudmesh.yaml")
        cometConf = config["cloudmesh.comet"]
        defaultUser = cometConf["username"]

        user = input("Comet nucleus username [%s]: " \
                     % defaultUser)
        if not user:
            user = defaultUser
        password = getpass.getpass()
        keyurl = "%s/getkey" % cometConf["endpoints"][endpoint]["nucleus_base_url"]
        headers = {"ACCEPT": "application/json"}
        r = requests.get(keyurl, headers=headers, auth=HTTPBasicAuth(user, password))
        if r.status_code == 200:
            keyobj = r.json()
            api_key = keyobj["key_name"]
            api_secret = keyobj["key"]
            config = ConfigDict("cloudmesh.yaml")
            config.data["cloudmesh"]["comet"]["endpoints"] \
                [endpoint]["auth_provider"] = 'apikey'
            config.data["cloudmesh"]["comet"]["endpoints"] \
                [endpoint]["apikey"]["api_key"] = api_key
            config.data["cloudmesh"]["comet"]["endpoints"] \
                [endpoint]["apikey"]["api_secret"] = api_secret

            config.save()
            Console.ok("api key retrieval and set was successful!")
        else:
            Console.error("Error getting api key. "
                          "Please check your username/password", traceflag=False)

def main():
    comet = Comet()

    # change user, password to proper value as set in django
    # in shell, we may ask user input

    print(comet.status())
    print(comet.logon())
    print(comet.status())
    print(comet.logoff())
    print(comet.status())
    print(comet.logoff())


def test_get_cluster_list():
    token = ''
    banner("TEST 1: Get without logon")
    authheader = {'content-type': 'application/json',
                  "Authorization": 'Token %s' % token}
    geturl = "https://localhost:8443/nucleus/v1/cluster/"
    r = requests.get(geturl, headers=authheader, verify=False)
    # pprint (r)
    pprint(r.json())

    banner("TEST 2: Auth and then get cluster list")
    # authurl = "https://localhost:8443/nucleus/rest-auth"
    # auth = Authenticator(authurl)
    # change user, password to proper value as set in django
    # in shell, we may ask user input
    comet = Comet()
    Comet.tunnelled = True
    token = comet.logon()

    # construct a header with auth token after login
    # for all the following calls before log out
    authheader = {'content-type': 'application/json',
                  "Authorization": 'Token %s' % token}
    geturl = "https://localhost:8443/nucleus/v1/"
    geturl1 = "{}cluster/".format(geturl)
    r = Comet.get(geturl1, headers=authheader)
    pprint(r)

    # as of 2:40pm ET Oct 15, this is changed to 'not implemented'
    # as of 5:30pm ET this is now fixed and working
    # Getting only cluster details for those owned by the caller.
    banner("TEST 3a: Get cluster 'OSG'")
    geturl1 = "%scluster/%s" % (geturl, "osg/")
    r1 = Comet.get(geturl1, headers=authheader)
    pprint(r1)

    banner("\nTEST 3b: Get cluster 'vc2' via tunnel")
    geturl1 = "%scluster/%s" % (geturl, "vc2/")
    r1 = Comet.get(geturl1, headers=authheader)
    pprint(r1)

    banner("\nTEST 3c: Get cluster 'vc2' directly")
    Comet.tunnelled = False
    geturl1 = Comet.url("cluster/vc2/")
    r1 = Comet.get(geturl1, headers=authheader)
    pprint(r1)

    banner("TEST 4: Get compute nodes sets")
    Comet.tunnelled = True
    r1 = Comet.get_computeset()
    pprint(r1)

    banner("TEST 4a: Get compute nodes set with id")
    r1 = Comet.get_computeset(46)
    pprint(r1)

    banner("TEST 10: logoff and get cluster list again")
    comet.logoff()
    authheader = {'content-type': 'application/json',
                  "Authorization": 'Token %s' % token}
    geturl = "https://localhost:8443/nucleus/v1/cluster/"
    r = requests.get(geturl, headers=authheader, verify=False)
    pprint(r.json())


def test_power_nodes(action='on'):
    banner("TEST: power on/off a list of nodes")

    banner("Authenticating...")
    # always logon first
    comet = Comet()
    Comet.tunnelled = True
    token = comet.logon()

    authheader = {'content-type': 'application/json',
                  "Authorization": 'Token %s' % token}

    url = "https://localhost:8443/nucleus/v1/"
    vcname = "vc2"
    vmnames = ["vm-vc2-0", "vm-vc2-1"]
    vmhosts = {
        vmnames[0]: "comet-01-05",
        vmnames[1]: "comet-01-06"
    }
    data = {"computes": [{"name": vm, "host": vmhosts[vm]} for vm in vmnames],
            "cluster": "%s" % vcname}

    if 'on' == action:
        banner("Issuing request to poweron nodes...")
        posturl = "{}/computeset/".format(url)
        # posturl = "%s%s/compute/poweron" % (url, vcname)
        r = Comet.http(posturl, action="post", headers=authheader, data=data)
        banner("RETURNED RESULTS:")
        print(r)
    elif 'off' == action:
        computesetid = 33
        banner("Issuing request to poweroff nodes...")
        puturl = "%s/computeset/%s/poweroff" % (url, computesetid)
        # posturl = "%s%s/compute/poweron" % (url, vcname)
        r = Comet.http(puturl, action="put", headers=authheader)
        banner("RETURNED RESULTS:")
        print(r)
    else:
        print("The Specified Power Action NOT Supported!")


if __name__ == "__main__":
    test_get_cluster_list()
    # main()
    test_power_nodes("off")
    test_power_nodes()



# from cloudmesh_client.hostlist import Parameter
import inspect
# from abc import ABCMeta

from cloudmesh_client.common.Printer import Printer
from cloudmesh_client.shell.console import Console


# noinspection PyBroadException,PyUnreachableCode,PyUnusedLocal
class CloudProviderBase(object):
    #    __metaclass__ = ABCMeta

    def __init__(self, cloudname, user=None, flat=False, source="db"):

        #
        # BUG libcloud_image does not belong here its only for libcloud_image not for other clouds.
        #
        self.kind = ["image", "flavor", "vm", "quota", "limits", "usage", "key"]
        self.nodes = None
        self.flavors = None
        self.data = None
        self.images = None
        self.quota = None
        self.limits = None
        self.usage = None
        self.cloudname = cloudname
        self.keys = None
        self.user = user
        self.secgroup = None
        self.credential = None
        self.driver = None
        self.flag = flat
        self.source = source

    def attributes(self, kind):
        """

        :param kind:
        :return: order and headers of the useful data
        """
        if kind in self.kind:
            header = None
            order = None
        else:
            raise ValueError("Kind " + kind + " not supported")

    def set_source(self, source):
        """
        Sets the source for the information to be returned. "db" and
        "cloud", "memory"
        :param source: the database can be queried in mode "db",
        the database can be bypassed in mode "cloud"
        """
        self.source = source
        if self.source in ["memory"]:
            raise ValueError("Memory source ot implemented yet.")
        return None

    # #########################
    # KIND MANAGEMENT
    # #########################

    def kinds(self, ):
        """
        returns a list of supported list and detail kinds
        :return: list of kinds supported
        :rtype: list
        """
        return self.kind

    def is_kind(self, name):
        """
        returns tru if the kind given by name exists
        :param name:
        :return:
        """
        return name in self.kind

    def add_kind(self, name):
        self.kind.append(name)

    def del_kind(self, name):
        self.kind.remove(name)

    def check_kind(self, name):
        """
        returns true if the kind given by name exists
        :param name:
        :return:
        """
        if not self.is_kind(name):
            raise ValueError("Kind " + name + " not supported")

    # #########################
    # RESOURCE
    # #########################

    def resource(self, function, kind, cloudname, **kwargs):
        """
        returns the objects in json format
        :param kind: the kind of list: vm, image, flavor, ...
        :param cloudname: if cloudname = none all cloudes, or cloudname = "all"
        :return:
        """
        """
        Listing of vm instances
        :return:
        """
        self.check_kind(kind)
        what = getattr(self, function + "_" + kind)
        return what(cloudname, **kwargs)

    def list(self, kind, cloudname, **kwargs):
        """
        returns the objects in json format
        :param kind: the kind of list: vm, image, flavor, ...
        :param cloudname: if cloudname = none all cloudes, or cloudname = "all"
        :return:
        """
        """
        Listing of vm instances
        :return:
        """
        # print (kwargs)
        return self.resource("list", kind, cloudname, **kwargs)

    def get(self, kind, cloudname, identifier, **kwargs):
        """
        Listing of vm instances
        :return:
        """
        return self.resource("get", kind, cloudname, **kwargs)

    def refresh(self, kind, cloudname, identifier, **kwargs):
        """
        Listing of vm instances
        :return:
        """
        return self.resource("refresh", kind, cloudname, **kwargs)

    # #########################
    # SECGROUP
    # #########################
    '''
    def list_secgroup(self, cloudname, **kwargs):
        """
        returns the objects in json format
        :param cloudname:
        :return:
        """
        """
        Listing of iamge
        :return:
        """
        raise ValueError(
            "{}: Not implemented yet.".format(inspect.stack()[0][3]))
        return None
    '''

    def list_secgroup_rules(self, cloudname):

        raise ValueError(
            "{}: Not implemented yet.".format(inspect.stack()[0][3]))
        return None

    # #########################
    # VMS
    # #########################

    def boot_vm(self, cloud, user, name, image, flavor, key, secgroup, meta, nics,
                *kwargs):
        """
        Boots a new vm instance on the target cloud.
        :param cloud:
        :param user:
        :param name:
        :param image:
        :param flavor:
        :param key:
        :param secgroup:
        :param meta:
        :return:
        """
        raise ValueError(inspect.stack()[0][3] + ": Not implemented yet.")
        return None

    def list_vm(self, cloudname, **kwargs):
        """
        returns the objects in json format
        :param cloudname:
        :return:
        """
        """
        Listing of vm instances
        :return:
        """
        raise ValueError(
            "{}: Not implemented yet.".format(inspect.stack()[0][3]))
        return None

    def get_vm(self, cloudname, identifier, **kwargs):
        """
        returns the objects in json format
        :param cloudname:
        :return:
        """
        """
        get vm instance
        :return:
        """
        raise ValueError(
            "{}: Not implemented yet.".format(inspect.stack()[0][3]))
        return None

    def refresh_vm(self, cloudname, identifier, **kwargs):
        """
        Listing of vm instances
        :return:
        """
        raise ValueError(
            "{}: Not implemented yet.".format(inspect.stack()[0][3]))
        return

    def rename_vm(self, current_name, new_name):
        """
        Renames a vm.
        :param current_name:
        :param new_name:
        :return:
        """
        raise ValueError(
            "{}: Not implemented yet.".format(inspect.stack()[0][3]))
        return

    def delete(self, name, group=None, force=None):
        """
        Deletes the vm indicated by name on target cloud.
        :param name:
        :param group:
        :param force:
        :return:
        """
        raise ValueError(
            "{}: Not implemented yet.".format(inspect.stack()[0][3]))
        return

    def get_ips(self, name, group=None, force=None):
        """
        Returns the ip addresses of the instance indicated by name
        :param name:
        :param group:
        :param force:
        :return:
        """
        raise ValueError(
            "{}: Not implemented yet.".format(inspect.stack()[0][3]))
        return

    # #########################
    # IMAGE
    # #########################

    def list_image(self, cloudname, **kwargs):
        """
        returns the objects in json format
        :param cloudname:
        :return:
        """
        """
        Listing of iamge
        :return:
        """
        raise ValueError(
            "{}: Not implemented yet.".format(inspect.stack()[0][3]))
        return None

    def get_image(self, **kwargs):
        """
        finds the image based on a query
        TODO: details TBD
        """
        raise ValueError(
            "{}: Not implemented yet.".format(inspect.stack()[0][3]))
        return None

    def refresh_image(self, cloudname, identifier, **kwargs):
        """
        Listing of vm instances
        :return:
        """
        raise ValueError(
            "{}: Not implemented yet.".format(inspect.stack()[0][3]))
        return

    # #########################
    # FLAVOR
    # #########################

    def list_flavor(self, cloudname, **kwargs):
        """
        returns the objects in json format
        :param cloudname:
        :return:
        """
        """
        Listing of iamge
        :return:
        """
        raise ValueError(
            "{}: Not implemented yet.".format(inspect.stack()[0][3]))
        return None

    def get_flavor(self, **kwargs):
        """
        finds the flavor based on a query
        TODO: details TBD
        """
        raise ValueError(
            "{}: Not implemented yet.".format(inspect.stack()[0][3]))
        return

    # noinspection PyUnusedLocal
    def refresh_flavor(self, cloudname, identifier, **kwargs):
        """
        Listing of vm instances
        :return:
        """
        raise ValueError(
            "{}: Not implemented yet.".format(inspect.stack()[0][3]))
        return

    # ####################################
    # KEYS
    # ####################################

    def add_key_to_cloud(self, name, public_key):
        """
        Adds key to cloud for given public key.
        :param name: Name of the keypair to create
        :param public_key: Existing public key string.
        :return:
        """
        """
        Creating public key.
        :return:
        """
        raise ValueError(
            "{}: Not implemented yet.".format(inspect.stack()[0][3]))
        return None

    def delete_key_from_cloud(self, name):
        """
        Deletes key from cloud for given name.
        :param name: Name of the keypair to create.
        :return:
        """
        """
        Creating public key.
        :return:
        """
        raise ValueError(
            "{}: Not implemented yet.".format(inspect.stack()[0][3]))
        return None

    def get_key(self, **kwargs):
        """
        finds the flavor based on a query
        TODO: details TBD
        """
        raise ValueError(
            "{}: Not implemented yet.".format(inspect.stack()[0][3]))
        return

    # noinspection PyUnusedLocal
    def refresh_key(self, cloudname, identifier, **kwargs):
        """
        Listing of vm instances
        :return:
        """
        raise ValueError(
            "{}: Not implemented yet.".format(inspect.stack()[0][3]))
        return

    def list_key(self, cloudname, **kwargs):
        """
        returns the objects in json format
        :param cloudname:
        :return:
        """
        """
        Listing of iamge
        :return:
        """
        raise ValueError(
            "{}: Not implemented yet.".format(inspect.stack()[0][3]))
        return None

    # ####################################
    # DETAILS
    # ####################################

    def details(self, kind, category, id, format="table"):

        from cloudmesh_client.db.CloudmeshDatabase import CloudmeshDatabase

        cm = CloudmeshDatabase()

        try:
            if kind not in self.kind:
                raise ValueError('{} not defined'.format(kind))

            elements = None
            for idkey in ["cm_id", "name", "uuid", "id", "cm_id"]:
                s = {idkey: id}
                try:
                    elements = cm.find(kind=kind, category=category, **s)
                except:
                    pass
                if elements is not None:
                    break

            if elements is None:
                return None

            if len(elements) > 0:
                element = elements[0]
                if format == "table":
                    return Printer.attribute(element)
                else:
                    return Printer.write(element,
                                         output=format)
            else:
                return None

        except Exception as ex:
            Console.error(ex.message)



""" run with

python setup.py install; nosetests -v --nocapture  tests/cm_cloud/test_vm.py:Test_vm.test_001

nosetests -v --nocapture

or

nosetests -v

"""

from cloudmesh_client.common.Shell import Shell
from cloudmesh_client.common.dotdict import dotdict
from cloudmesh_client.common.util import HEADING
from cloudmesh_client.common.util import banner
from cloudmesh_client.default import Default


# noinspection PyPep8Naming
class Test_vm:
    data = dotdict({
        "wrong_cloud": "no_cloud",
        "cloud": Default.cloud,
        "group": "test",
        "image": Default.get_image(category=Default.cloud),
        "flavor": Default.get_flavor(category=Default.cloud),
        "vm": "{}_testvm".format(Default.user),
        "vm_rename": "{}_renamed_testvm".format(Default.user),
    })
    data.image = Default.get_image()
    data.flavor = Default.get_flavor()

    def run(self, command):
        command = command.format(**self.data)
        banner(command, c="-")
        print(command)
        parameter = command.split(" ")
        shell_command = parameter[0]
        args = parameter[1:]
        result = Shell.execute(shell_command, args)
        print(result)
        return str(result)

    def setup(self):
        pass

    # noinspection PyPep8Naming
    def tearDown(self):
        pass

    def load_key(self):
        try:
            self.run("cm key load")
            result = self.run("cm key upload")
        except Exception as e:
            print(e)


    def test_000(self):
        HEADING("key upload")
        command= "cm key add --ssh"
        result = self.run(command)

    def test_001(self):
        HEADING("cm vm boot --name={vm} --cloud={cloud} --image={image} --flavor=2 --group={group}".format(**self.data))
        self.load_key()
        command = "cm vm boot --name={vm} --cloud={cloud} --image={image}" + \
                  " --flavor={flavor} --group={group}"
        result = self.run(command)
        assert "OK." in result

    def test_002(self):
        HEADING("cm vm refresh --cloud={cloud}".format(**self.data))
        result = self.run("cm vm refresh --cloud={cloud}")
        assert "OK." in result

    def test_003(self):
        HEADING("cm vm list --cloud={cloud}".format(**self.data))
        result = self.run("cm vm list --cloud={cloud}")
        assert "Listing VMs on Cloud: {cloud}".format(**self.data) in result

    def test_004(self):
        HEADING("cm vm list {vm} --cloud={cloud}".format(**self.data))
        result = self.run("cm vm list {vm} --cloud={cloud}")
        assert "Listing VMs on Cloud: {cloud}".format(**self.data) in result

    def test_005(self):
        HEADING("cm vm status --cloud={cloud}".format(**self.data))
        result = self.run("cm vm status --cloud={cloud}")
        assert "OK." in result

    def test_006(self):
        from pprint import pprint; pprint(self.data)
        HEADING("cm vm ip show {vm} --cloud={cloud}".format(**self.data))
        result = self.run("cm vm list --refresh --cloud={cloud}")

        result = self.run("cm vm ip show {vm} --cloud={cloud}")
        assert "name" in result
        assert "{vm}".format(**self.data) in result

    def test_007(self):
        HEADING("cm vm rename {vm} {vm_rename}".format(**self.data))
        result = self.run("cm vm rename {vm} {vm_rename}")
        assert "OK." in result

    def test_008(self):
        HEADING("cm vm delete {vm_rename} --cloud={cloud} ".format(**self.data))
        result = self.run("cm vm delete {vm_rename} --cloud={cloud}")
        assert "OK." in result

    def test_009(self):
        HEADING("testing purgeing a vm")
        result = self.run("cm key add --ssh")
        result = self.run("cm info")
        result = self.run("cm vm boot")
        result = self.run("cm vm list")
        result = self.run("cm vm delete last")
        result = self.run("cm vm list")

    '''
    def test_009(self):
        from pprint import pprint;
        pprint(self.data)
        HEADING("cm vm ip assign {vm} --cloud={cloud}".format(**self.data))
        result = self.run("cm vm list --refresh --cloud={cloud}")

        result = self.run("cm vm ip show {vm} --cloud={cloud}")
        assert "name" in result
        assert "{vm}".format(**self.data) in result
    '''



from __future__ import print_function

from cloudmesh_client.shell.console import Console
from cloudmesh_client.shell.command import command, PluginCommand, \
    CloudPluginCommand, ShellPluginCommand
from cloudmesh_client.default import Default
from cloudmesh_client.common.LogUtil import LogUtil

# get the logger
LOGGER = LogUtil.get_logger()


class LogLevelCommand(PluginCommand, CloudPluginCommand, ShellPluginCommand):
    topics = {"loglevel": "shell"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command loglevel")

    # noinspection PyUnusedLocal
    @command
    def do_loglevel(self, args, arguments):
        """
        ::

            Usage:
                loglevel set MODE [--cloud=CLOUD]
                loglevel get [--cloud=CLOUD]
                loglevel save [--cloud=CLOUD]

            Arguments:
                MODE    log level mode [DEBUG/INFO/WARNING/CRITICAL/ERROR]

            Options:
                --cloud=CLOUD    the name of the cloud

        Description:
            loglevel command sets the default logging level
            for a cloud.

        Examples:
            loglevel set DEBUG --cloud=kilo
                sets the default log level to DEBUG for kilo.

            loglevel get --cloud=kilo
                retreives the default log level for kilo cloud.

            loglevel save --cloud=kilo
                saves the log level preference to the db & yaml file.

        """
        # pprint(arguments)

        cloud = arguments["--cloud"] or Default.cloud
        LOGGER.info("Cloud: " + cloud + ", Arguments: " + str(arguments))

        if arguments["set"]:
            try:
                log_level = arguments["MODE"]
                response = LogUtil.set_level(log_level=log_level,
                                             cloudname=cloud)

                if response is not None:
                    Console.ok(response)
            except Exception as ex:
                Console.error(ex.message)

        elif arguments["get"]:
            try:
                log_level = LogUtil.get_level(cloudname=cloud)
                Console.ok("Current Log Level = " + log_level + ". Ok.")
            except Exception as ex:
                Console.error(ex.message)

        elif arguments["save"]:
            LogUtil.save(cloudname=cloud)
            pass



""" run with

python setup.py install; nosetests -v --nocapture  tests/cm_cloud/test_group.py:Test_group.test_001

nosetests -v --nocapture tests/test_group.py

or

nosetests -v tests/test_group.py

"""
from __future__ import print_function

from pprint import pprint

from cloudmesh_client.cloud.group import Group
from cloudmesh_client.common.Shell import Shell
from cloudmesh_client.common.dotdict import dotdict
from cloudmesh_client.common.util import HEADING
from cloudmesh_client.common.util import banner
from cloudmesh_client.default import Default


# noinspection PyPep8Naming
class Test_group:
    data = dotdict({
        "cloud": Default.cloud,
        "user": "test",
        "prefix": "test",
        "group": "groupA"
    })

    def run(self, command):
        command = command.format(**self.data)
        banner(command, c='-')
        print(command)
        parameter = command.split(" ")
        shell_command = parameter[0]
        args = parameter[1:]
        result = Shell.execute(shell_command, args)
        print(result)
        return result

    def setup(self):
        pass

    # noinspection PyPep8Naming
    def tearDown(self):
        pass


    def test_000(self):
        HEADING("testing cm group list")

        c = "cm group list"
        self.run(c)

    def test_001(self):
        HEADING("testing cm group add ")

        c = "cm group add {prefix}-001 --group={group}  --type=vm"
        self.run(c)

        index = str(Group.get(name="{group}".format(**self.data), scope='all'))
        print("INDEX", index)
        assert "test-001" in index

    def test_002(self):
        HEADING("testing cm group add")

        command = "cm group add {prefix}-002 --group={group}   --type=vm"
        self.run(command)

        group = Group.get(name="groupA", scope='all')

        print("GGGG", group)

        names = ["{prefix}-001".format(**self.data),
                 "{prefix}-002".format(**self.data)]
        for element in group:
            pprint(element)
            assert element["name"] == "groupA"
            assert element["species"] == "vm"
            assert element["member"] in names


    def test_003(self):
        HEADING("testing cm group copy groupA groupB")

        command = "cm group copy {group} groupB"

        result = self.run(command)
        assert "groupB" in result


    def test_004(self):
        HEADING("testing cm group merge groupA groupB groupC")
        command = "cm group merge {group} groupB groupC"

        result = self.run(command)
        assert "ok." in result

    def test_005(self):
        HEADING("testing cm group list  groupA")
        command = "cm group list  {group}"

        result = self.run(command)
        assert "groupA" in str(result)

    def test_006(self):
        HEADING("testing cm group list  --format=json groupA")
        command = "cm group list  --format=json {group}"

        result = self.run(command)
        assert "groupA" in result
        return

    def test_007(self):
        HEADING("testing cm group list  --format=table")
        command = "cm group list  --format=table"

        result = self.run(command)
        assert "groupA" in result

    def test_008(self):
        HEADING("testing cm group add with default cloud")
        banner("cm group add --name=groupX --id=albert-00x", c='-')

        result = self.run("cm default cloud")
        print("UUUU", result)
        pprint(self.data.cloud)
        print(type(result), type(self.data.cloud))
        assert self.data.cloud in result

        result = self.run("cm default type=vm")
        assert "ok." in result

        result = self.run("cm group add albert-00x --group=groupX ")
        # print ("RRR", result)
        # assert "albert-00x" in result

        result = self.run("cm group list groupX")

        print("DDDDD", result)

        assert "groupX" in result

    def test_009(self):
        HEADING("testing cm group remove ")
        banner("cm group remove {prefix}-002  --group={group}")

        result = self.run("cm group remove {prefix}-002  --group={group} ")
        print(result)
        assert "ok" in result

        result = self.run("cm group list {group}")
        assert "test-002" not in result

    def test_010(self):
        HEADING("testing cm group delete ")

        command = "cm group delete {group} "
        result = self.run(command)
        print(result)

        for group in ["groupA", "groupB", "groupC", "groupX"]:
            self.data.group = group
        command = "cm group delete {group} "

        result = self.run(command)

        assert "ok." in result

    def test_011(self):

        HEADING("list vms in group")

        setup = [
            "cm group add vm_1 --group=group_x",
            "cm group add vm_2 --group=group_x",
            "cm group add vm_3 --group=group_y",
            "cm group add vm_3 --group=group_x",
            "cm group list"]

        for command in setup:
            self.run(command)

        print("T1")
        result = Group.get_vms("group_x")
        print("group x", result)
        assert 'vm_1' in result
        assert 'vm_2' in result
        assert 'vm_3' in result

        print("T2")
        result = Group.get_vms("group_y")
        print("group y", result)
        assert 'vm_3' in result
        assert 'vm_1' not in result
        assert 'vm_2' not in result

        print("T3")
        result = Group.names()
        print("names", result)
        assert 'group_x' in result
        assert 'group_y' in result

        print("T4")
        result = Group.vm_groups("vm_1")
        print("vm_1", result)
        assert 'group_x' in result
        assert 'group_y' not in result

        print("T5")
        result = Group.vm_groups("vm_3")
        print("vm_3", result)
        assert 'group_x' in result
        assert 'group_y' in result

        print("T6")
        result = Group.delete(name="group_x")
        print(result)
        result = self.run("cm group list")
        print(result)

        assert 'vm_1' not in result
        assert 'vm_2' not in result

        print("T7")
        self.run("cm group list")

        for command in setup:
            self.run(command)
        result = self.run("cm group list")
        print(result)

        result = Group.delete(name="group_y")

        print(result)
        result = self.run("cm group list")

        assert 'group_y' not in result
        assert 'vm_3' in result

    def test_012(self):
        HEADING("list non existing group")
        command = "cm group list DOESNOTEXIST"
        result = self.run(command)

        assert "ERROR" in result



from __future__ import print_function
from cloudmesh_client.cloud.vc import Vc
from cloudmesh_client.shell.command import command, PluginCommand, CloudPluginCommand
from cloudmesh_client.shell.console import Console
from cloudmesh_client.default import Default
from cloudmesh_client.common.dotdict import dotdict
from cloudmesh_client.common.hostlist import Parameter
from pprint import pprint

class VcCommand(PluginCommand, CloudPluginCommand):
    topics = {"vc": "todo"}

    def __init__(self, context):
        self.context = context
        if self.context.debug:
            print("init command vc")

    # noinspection PyUnusedLocal
    @command
    def do_vc(self, args, arguments):
        """
        ::

            Usage:
                vc key add KEYFILE NAMES [--username=USERNAME] [--proxy=PROXY]
                vc key distribute NAMES [--username=USERNAME] [--proxy=PROXY]
                vc key list NAMES [--usort] [--username=USERNAME] [--proxy=PROXY] [--format=FORMAT]
                vc key proxy NAMES [--username=USERNAME] [--proxy=PROXY]

            Options:
               --format=FORMAT  the output format [default: table]

            Description:

                see examples

            Examples:
                cm vc key add keys.txt gregor-[001-010]
                    adds the keys in the file keys.txt to the authorized_keys file
                    in the user that is registered for the vm

                cm vc key add keys.txt gregor-[001-010] --username=ubuntu
                    adds the keys in the file keys.txt to the authorized_keys file
                    in the user ubuntu for each of the vms

                vc key distribute gregor-[001-010]
                    gathers the keys from the host gathers it into a single file
                    and adds them to the authorized keys file. Duplicated keys will
                    be ignored.

                vc key list gregor-[001-010] [--usort]
                    creates a table with all keys in authorized_keys from all of the
                    remote machines. If the parameter usort is specified it only lists
                    the key once, but lists in the host column the list of all host on
                    which the key is stored

                Proxy server

                    vc key proxy NAMES
                        sometimes you may not have enough floating IPs so it is possible to dedicate one machine
                        as a proxy server that has such a floating ip. The way this is done is that you need to set
                        up ssh tunnels via the proxy server in your  .ssh/config file. The command will print a
                        template that you could include in your .ssh/config file to gain easily access to your other
                        machines without floating ip. For example it will generate the following for a given PROXY host,
                        USERNAME, and vm1 is the name of the first vm in NAMES

                        Host vm1
                            User  USERNAME
                            Hostname PROXY
                            ProxyCommand  ssh 10.1.1.2 nc %h %p
                            ForwardX11 yes

                        Note: this is just a draft and will be improved upon discussion with the team

        """

        arg = dotdict(arguments)
        arg.usort = arguments["--usort"]
        arg.format = arguments["--format"]
        arg.username = arguments["--username"]
        arg.proxy = arguments["--proxy"]


        if arg.proxy:
            Console.error("proxy not yet supported", traceflag=False)

        if arg.NAMES is not None:
            arg.names = Parameter.expand(arg.NAMES)
        else:
            arg.names = None


        pprint (arg)

        if arg.add:

            print("vc key add KEYFILE NAMES --username=USERNAME")

            print(arg.username)
            print(arg.names)
            print(arg.KEYFILE)
            print(arg.proxy)


            Console.TODO("not yet implemented")
            return ""

        elif arg.distribute:

            print("vc key distribute NAMES --username=USERNAME")
            print(arg.names)
            print(arg.username)
            print(arg.proxy)

            Console.TODO("not yet implemented")
            return ""

        elif arg.list:

            print("vc key list NAMES [--usort] [--format=FORMAT]")

            print(arg.names)
            print(arg.username)
            print(arg.format)
            print(arg.usort)
            print(arg.proxy)
            result = Vc.list(names=arg.names)  # dont forget format and sort
            Console.TODO("not yet implemented")
            return ""

        elif arg.proxy:

            print("vc key proxy NAMES [--username=USERNAME] [--proxy=PROXY]")

            print(arg.names)
            print(arg.username)
            print(arg.format)
            print(arg.usort)
            print(arg.proxy)

            result = Vc.list(names=arg.names)  # dont forget format and sort
            Console.TODO("not yet implemented")
            return ""


""" run with


python setup.py install; nosetests -v --nocapture tests/cm_cloud/test_reservation.py:Test_reservation.test_001
python setup.py install; nosetests -v --nocapture  tests/cm_cloud/test_reservation.py

nosetests -v --nocapture

or

nosetests -v

"""

from cloudmesh_client.common.Shell import Shell
from cloudmesh_client.common.dotdict import dotdict
from cloudmesh_client.common.util import HEADING
from cloudmesh_client.common.util import banner
from cloudmesh_client.default import Default


# noinspection PyPep8Naming
class Test_reservation:
    data = dotdict({
        "cloud": Default.cloud,
        "json": "json",
        "yaml": "yaml",
        "user": "albert",
        "name": "test_name",
        "project": "cloudmesh",
        "wrong_cloud": "no_cloud",
        "category": Default.cloud
    })

    def run(self, command):
        command = command.format(**self.data)
        banner(command, c="-")
        print(command)
        parameter = command.split(" ")
        shell_command = parameter[0]
        args = parameter[1:]
        result = Shell.execute(shell_command, args)
        print(result)
        return str(result)

    def setup(self):
        pass

    # noinspection PyPep8Naming
    def tearDown(self):
        pass

    def test_001(self):
        HEADING(
            "cm reservation add --name=test "
            "--start='10/31/1988\ at\ 8:09\ pm' "
            "--end='10/21/2015\ at\ 9:00\ pm' "
            "--user=albert --project=cloudmesh "
            "--hosts=host001 --description=desc")
        result = self.run("cm reservation add --name={name} "
                          "--start='10/31/1988\ at\ 8:09\ pm' "
                          "--end='10/21/2015\ at\ 9:00\ pm' "
                          "--user={user} --project={project}")
        assert "OK." in result

    def test_002(self):
        HEADING("cm reservation list")
        result = self.run("cm reservation list")
        assert "OK." in result

    def test_003(self):
        HEADING("cm reservation list --user=albert")
        result = self.run("cm reservation list --user={user}")
        assert "OK." in result

    def test_004(self):
        HEADING("cm reservation list --user=albert --format=json")
        result = self.run("cm reservation list --user={user} --format={json}")
        assert "OK." in result

    def test_005(self):
        HEADING("cm reservation list --user=albert --format=yaml")
        result = self.run("cm reservation list --user={user} --format={yaml}")
        assert "OK." in result

    def test_006(self):
        HEADING("cm reservation update --name=test_name --project=another_proj")
        result = self.run("cm reservation update --name={name} --project=another_proj")
        assert "OK." in result

    def test_007(self):
        HEADING("cm reservation delete --name=test_name")
        result = self.run("cm reservation delete --name={name}")
        assert "OK." in result

    def test_008(self):
        HEADING("cm reservation delete all")
        result = self.run("cm reservation delete all")
        assert "OK." in result



from __future__ import print_function

from cloudmesh_client.shell.console import Console
from cloudmesh_client.common.Printer import Printer
from cloudmesh_client.db import CloudmeshDatabase
from cloudmesh_client.cloud.iaas.CloudProvider import CloudProvider

from cloudmesh_client.cloud.ListResource import ListResource
from cloudmesh_client.default import Default
from cloudmesh_client.common.ConfigDict import ConfigDict

from pprint import pprint

class Image(ListResource):
    cm = CloudmeshDatabase()

    @classmethod
    def refresh(cls, cloud):
        """
        This method would refresh the image list by first clearing
        the database, then inserting new data
        :param cloud: the cloud name
        """
        # Newly implemented refresh
        result = cls.cm.refresh("image", cloud)
        return result

    @classmethod
    def list(cls, cloud, format="table"):
        """
        This method lists all images of the cloud
        :param cloud: the cloud name
        """
        # TODO: make a CloudmeshDatabase without requiring the user=

        try:
            elements = cls.cm.find(kind="image", category=cloud, scope="all")

            (order, header) = CloudProvider(cloud).get_attributes("image")

            return Printer.write(elements,
                                 order=order,
                                 header=header,
                                 output=format)

        except Exception as ex:
            Console.error(ex.message)

    @classmethod
    def details(cls, cloud, id, live=False, format="table"):
        if live:
            cls.refresh(cloud)

        return CloudProvider(cloud).details('image', cloud, id, format)

    @classmethod
    def guess_username_from_category(cls, category, image, username=None):
        chameleon = "chameleon" in ConfigDict(filename="cloudmesh.yaml")["cloudmesh"]["clouds"][category]["cm_host"]
        username = None
        if chameleon:
            username = "cc"
        else:

            if username is None:
                Console.error("Could not guess the username of the vm", traceflag=False)
                return
            username = username or Image.guess_username(image)
        return username

    @classmethod
    def guess_username(cls, vm_name, cloud=None, description=None):
        username = None

        names = [vm_name]
        if description is not None:
            names.append(description)

        chameleon = cloud == "chameleon"
        for name in names:
            name = name.lower()
            if name.startswith("cc-") or chameleon:
                username = "cc"
                break
            elif any(x in name for x in ["ubuntu", "wily", "xenial"]):
                username = "ubuntu"
                break
            elif "centos" in name:
                username = "root"
                break
            elif "fedora" in name:
                username = "root"
                break
            elif "rhel" in name:
                username = "root"
                break
            elif "cirros" in name:
                username = "root"
                break
            elif "coreos" in name:
                username = "root"
                break

        return username

    @classmethod
    def get(cls, name=None, cloud=None):
        cloud = cloud or Default.cloud
        name = name or Default.image

        image = cls.cm.find(kind="image", category=cloud, name=name, output='dict', scope='first')
        return image

    @classmethod
    def get_username(cls, name, cloud, guess=False):
        image = cls.get(cloud=cloud, name=name)
        if guess and image.username is None:
            return cls.guess_username(image.name)
        return image.username

    @classmethod
    def set_username(cls, name=None, cloud=None, username=None):
        image = cls.get(cloud=cloud, name=name)

        cls.cm.set(name, "username", username, provider=image.provider, kind="image", scope="first")



import os
from pprint import pprint
from uuid import UUID

from cloudmesh_client.cloud.iaas.CloudProviderBase import CloudProviderBase
from cloudmesh_client.common.todo import TODO
from cloudmesh_client.common.ConfigDict import Config, ConfigDict
from cloudmesh_client.common.FlatDict import FlatDict
from keystoneclient.auth.identity import v3
from keystoneclient import session
from novaclient import client
import requests
import getpass
from cloudmesh_client.shell.console import Console


class CloudProviderCometAPI(CloudProviderBase):
    cloud_type = "openstack"

    def __init__(self, cloud_name, cloud_details, user=None, flat=True):
        super(CloudProviderCometAPI, self).__init__(cloud_name, user=user)
        self.flat = flat
        self.cloud_type = "comet"
        self.kind = ["image", "flavor", "vm", "key", "group"]

        self.provider = None
        self.default_image = None
        self.default_flavor = None
        self.cloud = None
        self.cloud_details = None

        self.initialize(cloud_name, cloud_details)

    def initialize(self, cloudname, user=None):

        d = ConfigDict("cloudmesh.yaml")

        """
        self.cloud_details = d["cloudmesh"]["clouds"][cloudname]

        # pprint(self.cloud_details)

        self.cloud = cloudname
        self.default_flavor = self.cloud_details["default"]["flavor"]
        self.default_image = self.cloud_details["default"]["image"]
        self.tenant = self.cloud_details['credentials']['OS_TENANT_NAME']
        version = 2
        credentials = self.cloud_details["credentials"]
        cert = False
        if "OS_CACERT" in credentials:
            if credentials["OS_CACERT"] is not False:
                cert = Config.path_expand(credentials["OS_CACERT"])
        auth_url = credentials["OS_AUTH_URL"]
        ksversion = auth_url.split("/")[-1]
        """

        """
        # GitHub issue 101
        # mechanism to interactively ask for password
        # when OS_PASSWORD set as "readline",
        # or read os.environ if set as "env".
        os_password = credentials["OS_PASSWORD"]
        if os_password.lower() in ["readline", "read", "tbd"]:
            os_password = getpass.getpass()
        elif os_password.lower() == "env":
            os_password = os.environ.get("OS_PASSWORD", getpass.getpass())

        if "v2.0" == ksversion:
            self.provider = client.Client(
                version,
                credentials["OS_USERNAME"],
                os_password,
                credentials["OS_TENANT_NAME"],
                credentials["OS_AUTH_URL"],
                cert)
        elif "v3" == ksversion:
            sess = session.Session(auth=self._ksv3_auth(credentials),
                                   verify=cert)
            self.provider = client.Client(2, session=sess)
        """

    def list_flavor(self, cloudname, **kwargs):
        # d = self._to_dict(self.provider.flavors.list())
        return d

    def list_image(self, cloudname, **kwargs):
        # d = self._to_dict(self.provider.images.list())
        return d

    def list_vm(self, cloudname, **kwargs):

        return None

    def rename_vm(self, current_name, new_name):
        pass

    def list_ips(self, **kwargs):
        pass

    def list_console(self, name, length=None):
        server = self.provider.servers.get(name)
        log = server.get_console_output(length=None)
        return log

    def boot_vm(self,
                name,
                group=None,
                image=None,
                flavor=None,
                cloud=None,
                key=None,
                secgroup=None,
                meta=None,
                nics=None,
                **kwargs):
        """
        Spawns a VM instance on the cloud.
        If image and flavor passed as none, it would consider the defaults specified in cloudmesh.yaml.

        :param name: Name of the instance to be started
        :param image: Image id to be used for the instance
        :param flavor: Flavor to be used for the instance
        :param cloud: Cloud on which to spawn the machine. Defaults to 'India'.
        :param key: Key to be used for the instance
        :param secgroup: Security group for the instance
        :param meta: A dict of arbitrary key/value metadata to store for this server
        """

        return None

    def delete_vm(self, name, group=None, force=None):
        """
        Deletes a machine on the target cloud indicated by the id or name
        :param name: Name or ID of the instance to be deleted
        :param group: Security group of the instance to be deleted
        :param force: Force delete option
        :return:
        """
        return None

    def start_vm(self, name, group=None, force=None):
        """
        Starts a suspended machine on the target cloud indicated by the id or name
        :param name: Name or ID of the instance to be deleted
        :param group: Security group of the instance to be deleted
        :param force: Force delete option
        :return:
        """

        if self.isUuid(name):
            server = self.provider.servers.get(name)
        else:
            server = self.provider.servers.find(name=name)

        server.start()

    def stop_vm(self, name, group=None, force=None):
        """
        Stops a machine on the target cloud indicated by the id or name
        :param name: Name or ID of the instance to be deleted
        :param group: Security group of the instance to be deleted
        :param force: Force delete option
        :return:
        """

        if self.isUuid(name):
            server = self.provider.servers.get(name)
        else:
            server = self.provider.servers.find(name=name)

        server.stop()

    def get_ips(self, name, group=None, force=None):
        """
        Returns the ip of the instance indicated by name
        :param name:
        :param group:
        :param force:
        :return: IP address of the instance
        """
        if self.isUuid(name):
            server = self.provider.servers.get(name)
        else:
            server = self.provider.servers.find(name=name)

        return self.provider.servers.ips(server)

    def add_key(self, name, public_key):
        pass

    def delete_key(self, name):
        pass

    def get_image(self, **kwargs):
        pass


    def get_ip(self, **kwargs):
        pass

    def get_flavor(self, **kwargs):
        pass

    def get_vm(self, **kwargs):
        pass

    def attributes(self, kind):

        layout = {}

        if kind in layout:
            order = layout[kind]["order"]
            header = layout[kind]["header"]
        else:
            order = None
            header = None

        return order, header


# CloudProviderBase.register(CloudProviderOpenstackAPI)

if __name__ == "__main__":
    cloudname = 'kilo'
    d = ConfigDict("cloudmesh.yaml")
    cloud_details = d["cloudmesh"]["clouds"][cloudname]

    cp = CloudProviderCometAPI(cloudname, cloud_details)

    d = {'name': '390792c3-66a0-4c83-a0d7-c81e1c787710'}
    pprint(cp.list_quota(cloudname))

    pprint(cp.list_key(cloudname))




from __future__ import print_function

from cloudmesh_client.shell.console import Console
from cloudmesh_client.common.Printer import Printer
from cloudmesh_client.db import CloudmeshDatabase
from cloudmesh_client.cloud.iaas.CloudProvider import CloudProvider
from cloudmesh_client.cloud.ListResource import ListResource


'''


stdin, stdout, stderr = Shell.kegyen('ssh-keygen -t rsa -f id_rsa.key')
stdin, stdout, stderr = Shell.cat('id_rsa.key.pub')


class DistributeKeys(object):

    def read_ips(filename):

fo = open(<IP_LIST_FILE>, "rw+")

pubkeys = []


for ln in fo:
    ip = ln.split(",")
    usr = ip[0]
    addr = ip[1].replace('\n', '')
    op =stdout.read().splitlines()
    pubkeys.append(op[0])



full_keylst = open("team_pub_keys.txt","a")
for k in pubkeys:
    full_keylst.write(k+"\n")

fo = open("ips.txt", "rw+")
for ln in fo:
    ip = ln.split(",")
    usr = ip[0]
    addr = ip[1].replace('\n', '')
    print "Connecting to ",addr
    sftp = ssh.open_sftp()
    f = sftp.open("~/.ssh/authorized_keys",'w')
    keys = open("team_pub_keys.txt","r")
    for key in keys:
        f.write(key)
    print "Public keys written into",addr
    f.close()
    ssh.close()

'''


class Vc(ListResource):
    cm = CloudmeshDatabase()


    '''
    vc key list NAMES [--usort]
    '''

    @classmethod
    def list(cls, names, usort=False, format="table"):
        """
        This method lists all vcs of the cloud
        :param cloud: the cloud name
        """

        Console.error("NOT YET IMPLEMENTED")
        return None







""" run with

python setup.py install; nosetests -v --nocapture  tests/cm_libcloud/test_libcloud_api.py:Test_libcloud_native.test_001

nosetests -v --nocapture tests/libcloud/test_libcloud_api.py

or

nosetests -v tests/test_image.py

"""

from __future__ import print_function

import re
from pprint import pprint

from libcloud.compute.providers import get_driver
from libcloud.compute.types import Provider

from cloudmesh_client.common.ConfigDict import ConfigDict
from cloudmesh_client.common.util import HEADING


# noinspection PyPep8Naming
class Test_libcloud_native(object):
    """
    Tests the libcloud native connection to chameleon
    """

    driver = None
    credential = None
    clouddefault = None

    def setup(self):
        cloud = "cybera-ec2"
        config = ConfigDict("cloudmesh.yaml")
        self.credential = config['cloudmesh']['clouds'][cloud]['credentials']
        self.clouddefault = config['cloudmesh']['clouds'][cloud]['default']
        # pprint(dict(self.credential))

        auth_url = self.credential["EC2_URL"]

        data = re.match(r'^http[s]?://(.+):([0-9]+)/([a-zA-Z/]*)',
                        auth_url,
                        re.M | re.I)
        host, port, path = data.group(1), data.group(2), data.group(3)
        print("host: " + host)
        print("port: " + port)
        print("path: " + path)

        extra_args = {'path': path}
        cls = get_driver(Provider.EC2_US_EAST)
        self.driver = cls(
            self.credential['EC2_ACCESS_KEY'],
            self.credential['EC2_SECRET_KEY'],
            host=host,
            port=port,
            **extra_args)
        print("DRIVER", self.driver)
        assert "libcloud.compute.drivers.ec2.EC2NodeDriver object at" in str(self.driver)

    def test_002(self):
        """list VMs"""
        HEADING()
        pprint(self.driver)
        nodes = self.driver.list_nodes()
        pprint("Nodes", nodes)
        assert True

    def test_003(self):
        """list images"""
        HEADING()
        print(self.driver)
        images = self.driver.list_images()
        print("Images", images)
        assert len(images) > 0

    def test__004(self):
        """list flavors"""
        HEADING()
        print(self.driver)
        sizes = self.driver.list_sizes()
        print("Flavor", sizes[0])
        assert len(sizes) > 0

    def test_005(self):
        # specify flavor and image
        HEADING()
        print(self.driver)
        myflavor = self.clouddefault['flavor']
        sizes = self.driver.list_sizes()
        size = [s for s in sizes if s.id == myflavor][0]
        print(size)
        assert True

    def test_006(self):
        # Changed "name" -> "id" (diff from openstack)
        HEADING()
        print(self.driver)
        myimage = self.clouddefault['image']
        images = self.driver.list_images()
        image = [i for i in images if i.name == myimage][0]
        print(image)
        assert True

    def test_007(self):
        """launch a new VM"""
        HEADING()
        print(self.driver)

        myimage = self.clouddefault['image']
        images = self.driver.list_images()
        image = [i for i in images if i.name == myimage][0]

        myflavor = self.clouddefault['flavor']
        sizes = self.driver.list_sizes()
        size = [s for s in sizes if s.id == myflavor][0]

        name = "{:}-libcloud".format(self.credential['userid'])
        node = self.driver.create_node(name=name, image=image, size=size)
        print(node)
        assert True

    def test_008(self):
        """check if the new VM is in the list"""
        HEADING()
        print(self.driver)
        nodes = self.driver.list_nodes()
        print(nodes)
        assert True

    '''
    def test_009(self):
        """public ip"""
        HEADING()
        # wait the node to be ready before assigning public IP
        sleep(10)

        # public IPs
        # get the first pool - public by default
        # create an ip in the pool
        elastic_ip = self.driver.ex_allocate_address()

        # attach the ip to the node
        self.driver.ex_associate_address_with_node(node, elastic_ip)

        # check updated VMs list to see if public ip is assigned
        nodes = self.driver.list_nodes()
        print (nodes)
        assert True


    def test_010(self):
        """remove node"""
        HEADING()
        # delete the ip

        # delete vm
        node.destroy()
        assert True
    '''



import os
from pprint import pprint
from uuid import UUID

from cloudmesh_client.cloud.iaas.CloudProviderBase import CloudProviderBase
from cloudmesh_client.common.todo import TODO
from cloudmesh_client.common.ConfigDict import Config, ConfigDict
from cloudmesh_client.common.FlatDict import FlatDict
from keystoneclient.auth.identity import v3
from keystoneclient import session
from novaclient import client
import requests
import getpass
from cloudmesh_client.shell.console import Console

requests.packages.urllib3.disable_warnings()


# TODO: unset as not allowing to smoothly switch
def set_os_environ(cloudname):
    """Set os environment variables on a given cloudname"""
    # TODO: this has a severe bug as it is not unsetting variables
    # Also this coded duplicates in part from register
    try:
        d = ConfigDict("cloudmesh.yaml")
        credentials = d["cloudmesh"]["clouds"][cloudname]["credentials"]
        for key, value in list(credentials.items()):
            if key == "OS_CACERT":
                os.environ[key] = Config.path_expand(value)
            else:
                os.environ[key] = value
    except Exception as e:
        Console.error("problem setting env")


#
# we already have a much better convert to dict function
#


# noinspection PyPep8Naming,PyUnusedLocal,PyUnusedLocal
class CloudProviderOpenstackAPI(CloudProviderBase):
    cloud_type = "openstack"
    cloud_pwd = {}

    def __init__(self, cloud_name, cloud_details, user=None, flat=True):
        super(CloudProviderOpenstackAPI, self).__init__(cloud_name, user=user)
        self.flat = flat
        self.cloud_type = "openstack"
        self.kind = ["image", "flavor", "vm", "quota", "limits", "usage", "key", "group"]
        self.dbobject = ["image", "flavor", "vm", "quota", "limits", "usage", "key", "group"]

        self.provider = None
        self.default_image = None
        self.default_flavor = None
        self.cloud = None
        self.cloud_details = None

        self.initialize(cloud_name, cloud_details)

    def _to_dict(self, openstack_result):
        d = {}
        for index, value in enumerate(openstack_result):
            # print "---"
            d[index] = dict(value.__dict__["_info"])
            if 'links' in d[index]:
                del d[index]['links']
            if 'server__links' in d[index]:
                del d[index]['server__links']
            if 'image__links' in d[index]:
                del d[index]['image__links']
            if 'flavor__links' in d[index]:
                del d[index]['flavor__links']

            # If flat dict flag set, convert to flatdict
            if self.flat is True:
                d[index] = dict(FlatDict(d[index]).dict)
            else:
                ValueError("TODO: we only support flat for now")
        return d

    def _ksv3_auth(self, credentials):
        # auth to identity v3

        # ref: http://docs.openstack.org/developer/python-keystoneclient/api/keystoneclient.auth.identity.v3.html#keystoneclient.auth.identity.v3.Password
        allowed_params = ["auth_url",
                          "password",
                          "username",
                          "user_id",
                          "user_domain_id",
                          "user_domain_name",
                          "trust_id",
                          "domain_id",
                          "domain_name",
                          "project_id",
                          "project_name",
                          "project_domain_id",
                          "project_domain_name"]
        authdict = {}
        # always required
        authdict["auth_url"] = credentials["OS_AUTH_URL"]
        authdict["password"] = credentials["OS_PASSWORD"]

        # setting automatically all available ones
        # CAUTION: MAY be causing conflict/error.
        # e.g., for jetstream, the OS_USER_DOMAIN_ID=tacc
        # was causing error (domain not found)
        for key in credentials:
            if key.startswith("OS_"):
                newkey = key[3:].lower()
                if newkey in allowed_params:
                    authdict[newkey] = credentials[key]

        '''
        # different cloud provider may be using different set of other info
        # e.g. id or name for project, domain, etc.
        if "OS_USERNAME" in credentials:
            authdict["username"] = credentials["OS_USERNAME"]
        if "OS_USER_ID" in credentials:
            authdict["user_id"] = credentials["OS_USER_ID"]
        if "OS_PROJECT_NAME" in credentials:
            authdict["project_name"] = credentials["OS_PROJECT_NAME"]
        if "OS_PROJECT_ID" in credentials:
            authdict["project_id"] = credentials["OS_PROJECT_ID"]
        if "OS_USER_DOMAIN_NAME" in credentials:
            authdict["user_domain_name"] = credentials["OS_USER_DOMAIN_NAME"]
        elif "OS_USER_DOMAIN_ID" in credentials:
            authdict["user_domain_name"] = credentials["OS_USER_DOMAIN_ID"]
        if "OS_PROJECT_DOMAIN_NAME" in credentials:
            authdict["project_domain_name"] = credentials["OS_PROJECT_DOMAIN_NAME"]
        elif "OS_PROJECT_DOMAIN_ID" in credentials:
            authdict["project_domain_name"] = credentials["OS_PROJECT_DOMAIN_ID"]
        '''
        # pprint(authdict)

        ksauth = v3.Password(**authdict)

        return ksauth

    def logon(self, cloudname):
        """
        Logs onto a cloud
        :param cloudname:
        :return:
        """
        CloudProviderOpenstackAPI.cloud_pwd[cloudname] = {}
        d = ConfigDict("cloudmesh.yaml")
        credentials = d["cloudmesh"]["clouds"][cloudname]["credentials"]

        os_password = credentials["OS_PASSWORD"]
        prompt = "Password for cloud - {}:".format(cloudname)
        if os_password.lower() == "readline":
            if 'pwd' in CloudProviderOpenstackAPI.cloud_pwd[cloudname]:
                os_password = CloudProviderOpenstackAPI.cloud_pwd[cloudname]["pwd"]
            else:
                os_password = getpass.getpass(prompt=prompt)
        elif os_password.lower() == "env":
            if 'pwd' in CloudProviderOpenstackAPI.cloud_pwd[cloudname]:
                os_password = CloudProviderOpenstackAPI.cloud_pwd[cloudname]["pwd"]
            else:
                os_password = os.environ.get("OS_PASSWORD", getpass.getpass(prompt=prompt))

        #
        # TODO: pwd is not standing for passwd
        #
        CloudProviderOpenstackAPI.cloud_pwd[cloudname]["pwd"] = os_password
        CloudProviderOpenstackAPI.cloud_pwd[cloudname]["status"] = "Active"

        return

    def activate(self, cloudname):
        """
        Activates a cloud
        :param cloudname:
        :return:
        """
        CloudProviderOpenstackAPI.cloud_pwd[cloudname]["status"] = "Active"
        return

    def deactivate(self, cloudname):
        """
        Deactivates a cloud
        :param cloudname:
        :return:
        """
        CloudProviderOpenstackAPI.cloud_pwd[cloudname]["status"] = "Inactive"
        return

    def logout(self, cloudname):
        """
        Logs out from a cloud
        :param cloudname:
        :return:
        """
        CloudProviderOpenstackAPI.cloud_pwd.pop(cloudname)
        return

    def list_clouds(self):
        """
        Lists clouds
        :return:
        """
        cloud_dict = {}
        d = ConfigDict("cloudmesh.yaml")
        clouds = list(d["cloudmesh"]["clouds"].keys())
        for i, cloud in enumerate(clouds):
            if cloud in CloudProviderOpenstackAPI.cloud_pwd:
                cloud_dict[i] = {
                    "cloud": cloud,
                    "status": CloudProviderOpenstackAPI.cloud_pwd[cloud]["status"]
                }
            else:
                cloud_dict[i] = {
                    "cloud": cloud,
                    "status": "Logged Out"
                }

        return cloud_dict

    def initialize(self, cloudname, user=None):

        d = ConfigDict("cloudmesh.yaml")
        self.cloud_details = d["cloudmesh"]["clouds"][cloudname]

        # pprint(self.cloud_details)

        self.cloud = cloudname
        self.default_flavor = self.cloud_details["default"]["flavor"]
        self.default_image = self.cloud_details["default"]["image"]
        self.tenant = self.cloud_details['credentials']['OS_TENANT_NAME']
        version = 2
        credentials = self.cloud_details["credentials"]
        cert = False
        if "OS_CACERT" in credentials:
            if credentials["OS_CACERT"] is not False:
                cert = Config.path_expand(credentials["OS_CACERT"])
        auth_url = credentials["OS_AUTH_URL"]
        ksversion = auth_url.split("/")[-1]

        """
        # GitHub issue 101
        # mechanism to interactively ask for password
        # when OS_PASSWORD set as "readline",
        # or read os.environ if set as "env".
        """
        os_password = credentials["OS_PASSWORD"]
        prompt = "Password for cloud - {}:".format(cloudname)
        if os_password.lower() in ["readline", "read", "tbd"]:
            if cloudname in CloudProviderOpenstackAPI.cloud_pwd and \
                        'pwd' in CloudProviderOpenstackAPI.cloud_pwd[cloudname]:
                os_password = CloudProviderOpenstackAPI.cloud_pwd[cloudname]["pwd"]
            else:
                os_password = getpass.getpass(prompt=prompt)
        elif os_password.lower() == "env":
            if cloudname in CloudProviderOpenstackAPI.cloud_pwd and \
                        'pwd' in CloudProviderOpenstackAPI.cloud_pwd[cloudname]:
                os_password = CloudProviderOpenstackAPI.cloud_pwd[cloudname]["pwd"]
            else:
                os_password = os.environ.get("OS_PASSWORD", getpass.getpass(prompt=prompt))

        # store the password for this session
        CloudProviderOpenstackAPI.cloud_pwd[cloudname] = {}
        CloudProviderOpenstackAPI.cloud_pwd[cloudname]["pwd"] = os_password
        CloudProviderOpenstackAPI.cloud_pwd[cloudname]["status"] = "Active"

        if "v2.0" == ksversion:
            kws = dict(
                username = credentials['OS_USERNAME'],
                api_key  = os_password,
                project_id = credentials['OS_TENANT_NAME'],
                auth_url = credentials['OS_AUTH_URL'],
            )
            if cert:
                kws['cacert'] = cert
                if not os.path.exists(cert):
                    Console.error('Certificate file `{}` not found'.format(cert))
            self.provider = client.Client(version, **kws)
        elif "v3" == ksversion:
            sess = session.Session(auth=self._ksv3_auth(credentials),
                                   verify=cert)
            self.provider = client.Client(2, session=sess)

    def mode(self, source):
        """
        Sets the source for the information to be returned. "db" and "cloud"
        :param source: the database can be queried in mode "db",
        the database can be bypassed in mode "cloud"
        """
        TODO.implement()

    def list_key(self, cloudname, **kwargs):

        # TODO: this needs to be move to the provider
        _keys = self._to_dict(self.provider.keypairs.list())

        for id in _keys:
            key = _keys[id]

            # key["type"], key["string"], key["comment"] = (key["keypair__public_key"] + " ").split(" ", 3)
            # key["comment"] = key["comment"].strip()
            key_segments = key["keypair__public_key"].split(" ")
            key["type"] = key_segments[0]
            key["string"] = key_segments[1]
            key["comment"] = ''
            if len(key_segments) == 3:
                key["comment"] = key_segments[2]
            elif len(key_segments) > 3:
                key["comment"] = " ".join(key_segments[2:])

        return _keys
        # return self._to_dict(self.provider.keypairs.list())

    def list_flavor(self, cloudname, **kwargs):
        d = self._to_dict(self.provider.flavors.list())
        return d

    def list_image(self, cloudname, **kwargs):
        d = self._to_dict(self.provider.images.list())
        for e in d:
            o = d[e]
            if 'server__links' in o:
                del o['server__links']
        return d

    def list_secgroup(self, cloudname, **kwargs):
        return self._to_dict(self.provider.security_groups.list())

    def list_secgroup_rules(self, cloudname):

        groups = self.list_secgroup(cloudname)

        rules = []

        for id in groups:
            group = groups[id]
            for rule in group["rules"]:

                if rule['ip_protocol'] is not None:

                    element = {
                        'fromPort': rule["from_port"],
                        'toPort': rule["to_port"],
                        'group': group["name"],
                        'protocol': rule['ip_protocol'],
                        'ruleid': rule['id'],
                        'groupid': rule['parent_group_id']
                    }

                    if 'cidr' in rule['ip_range']:
                        element['ipRange'] = rule['ip_range']['cidr']
                    else:
                        element['ip_range'] = None
                    rules.append(element)
        return rules

    def create_secgroup(self, secgroup_name):
        secgroup = self.provider.security_groups \
            .create(name=secgroup_name,
                    description="Security group {}".format(secgroup_name))

        return secgroup

    def add_secgroup_rule(self, **kwargs):
        rule_id = self.provider.security_group_rules.create(kwargs["uuid"],
                                                            ip_protocol=kwargs["protocol"],
                                                            from_port=kwargs["from_port"],
                                                            to_port=kwargs["to_port"],
                                                            cidr=kwargs["cidr"])
        return rule_id

    def delete_secgroup(self, name):
        ret = None
        groups = self.provider.security_groups.list()

        if groups is not None:
            for sec_group in groups:

                # delete the secgroup in the cloud
                if sec_group.name == name:
                    try:
                        ret = self.provider.security_groups.delete(sec_group)
                        Console.msg("Secgroup delete: {} {}".format(sec_group.name, sec_group.__dict__["id"]))
                    except Exception as e:
                        Console.error(e.message, traceflag=False)
                        Console.error("Secgroup delete: {}".format(sec_group.name), traceflag=False)

        else:
            print("Could not find security group [{}] in cloud [{}]"
                  .format(name, self.cloud))
        return ret

    def delete_secgroup_rule(self, rule_id):
        return self.provider.security_group_rules.delete(rule_id)
        #return

    def list_vm(self, cloudname, **kwargs):
        vm_dict = self._to_dict(self.provider.servers.list())

        # Extracting and modifying dict with IP details as per the model requirement.
        ip_detail_dict = dict()
        secgroup_list_dict = dict()
        for index in vm_dict:
            ip_detail_dict[index] = dict()
            ip_detail_dict[index][u"floating_ip"] = None
            ip_detail_dict[index][u"static_ip"] = None
            for key in vm_dict[index]:
                if key.startswith("addresses"):
                    for ip_detail in vm_dict[index][key]:
                        if ip_detail["OS-EXT-IPS:type"] is not None:
                            if ip_detail["OS-EXT-IPS:type"] == "fixed":
                                ip_detail_dict[index][u"static_ip"] = ip_detail["addr"]
                            elif ip_detail["OS-EXT-IPS:type"] == "floating":
                                ip_detail_dict[index][u"floating_ip"] = ip_detail["addr"]

            secgroup_list_dict[index] = ""

            sec_index = 0
            if "security_groups" in vm_dict[index]:
                for secgroup in vm_dict[index][u"security_groups"]:
                    secgroup_list_dict[index] += secgroup["name"]
                    sec_index += 1
                    # Append comma if not the last element
                    if sec_index < len(vm_dict[index]["security_groups"]):
                        secgroup_list_dict[index] += ","

        for index in vm_dict:
            vm_dict[index][u"floating_ip"] = ip_detail_dict[index]["floating_ip"]
            vm_dict[index][u"static_ip"] = ip_detail_dict[index]["static_ip"]
            vm_dict[index][u"security_ groups"] = secgroup_list_dict[index]

        for e in vm_dict:
            o = vm_dict[e]
            for link in ['server_links', 'flavor__links', 'image__links']:
                if link in o:
                    del o[link]

        return vm_dict

    def rename_vm(self, current_name, new_name):
        """
        Renames a vm.
        :param current_name:
        :param new_name:
        :return:
        """
        if self.isUuid(current_name):
            server = self.provider.servers.get(current_name)
            print("Renaming VM ({})".format(current_name))
            server.update(name=new_name)
        else:
            # server = self.provider.servers.find(name=name)
            search_opts = {
                'name': current_name,
            }
            vms = self.provider.servers.list(search_opts=search_opts)
            for vm in vms:
                print("Renaming VM ({}) : {}".format(vm.name, vm.id))
                vm.update(name=new_name)

    def list_limits(self, tenant, **kwargs):
        return self.provider.limits.get(tenant_id=tenant).__dict__["_info"]

    def list_floating_ips(self, **kwargs):
        return self._to_dict(self.provider.floating_ips.list())

    def list_floating_ip_pools(self, **kwargs):
        return self._to_dict(self.provider.floating_ip_pools.list())

    def list_quota(self, cloudname, **kwargs):
        # pprint(self.provider.__dict__)
        # pprint(dir(self.provider.quotas))
        tenant = self.cloud_details["credentials"]["OS_TENANT_NAME"]
        # print (tenant)
        # pprint(self.provider.quotas.defaults(tenant))
        result = self.provider.quotas.defaults(tenant).__dict__["_info"]
        del result['id']
        return result

    def list_usage(self, cloudname, **kwargs):
        raise ValueError("list usage is not supported")

    def list_console(self, name, length=None):
        server = self.provider.servers.get(name)
        log = server.get_console_output(length=None)
        return log

    def boot_vm(self,
                name,
                group=None,
                image=None,
                flavor=None,
                cloud=None,
                key=None,
                secgroup=None,
                meta=None,
                nics=None,
                **kwargs):
        """
        Spawns a VM instance on the cloud.
        If image and flavor passed as none, it would consider the defaults specified in cloudmesh.yaml.

        :param name: Name of the instance to be started
        :param image: Image id to be used for the instance
        :param flavor: Flavor to be used for the instance
        :param cloud: Cloud on which to spawn the machine. Defaults to 'India'.
        :param key: Key to be used for the instance
        :param secgroup: Security group for the instance
        :param meta: A dict of arbitrary key/value metadata to store for this server
        """

        if cloud is None:
            Console.error("Cloud is not specified")
            return None

        image = image or self.default_image
        flavor = flavor or self.default_flavor

        image_id = self.get_image_id(image)
        flavor_id = self.get_flavor_id(flavor)

        # if no nics specified, try to find one in case it's needed
        if nics is None or len(nics) == 0 or (len(nics) == 1 and nics[0]['net-id'] is None):
            # get net-id based on tenant network name, and other possible
            # default network names
            netnames = ["{0}-net".format(self.tenant), "int-net"]
            nicsmy = self.provider.networks.list()
            for netname in netnames:
                nic = [x for x in nicsmy if x.label == netname]
                if len(nic) > 0:
                    # found and break out of loop
                    # use the value found here
                    nic = nic[0]
                    nics = [{"net-id": nic.id}]
                    break

        """
        create(name, image, flavor, meta=None, files=None, reservation_id=None, min_count=None, max_count=None,
          security_groups=None, userdata=None, key_name=None, availability_zone=None, block_device_mapping=None,
          block_device_mapping_v2=None, nics=None, scheduler_hints=None, config_drive=None, disk_config=None,
          admin_pass=None, access_ip_v4=None, access_ip_v6=None, **kwargs)
        """
        if meta['cluster'] is None:
            meta['cluster'] = 'None'
        d = {
            "name": name,
            "image_id": image_id,
            "flavor_id": flavor_id,
            "meta": meta,
            "key_name": key,
            "security_groups": secgroup,
            "nics": nics}
        pprint (d)
        id = None
        try:

            server = self.provider.servers.create(name,
                                                  image_id,
                                                  flavor_id,
                                                  meta=meta,
                                                  key_name=key,
                                                  security_groups=secgroup,
                                                  nics=nics)

            # return the server id
            id = server.__dict__["id"]
            return id
        except Exception as e:
            if "Invalid key_name provided." in str(e):
                Console.error("Invalid key provided. "
                              "Is the key loaded, the default key set properly and uploaded to the cloud?",
                              traceflag=False)
            else:
                Console.error("Problem starting vm", traceflag=False)
                Console.error(e.message)

        return id

    def delete_vm(self, name, group=None, force=None):
        """
        Deletes a machine on the target cloud indicated by the id or name
        :param name: Name or ID of the instance to be deleted
        :param group: Security group of the instance to be deleted
        :param force: Force delete option
        :return:
        """

        if self.isUuid(name):
            server = self.provider.servers.get(name)
        else:
            # server = self.provider.servers.find(name=name)
            search_opts = {
                'name': name,
            }
            vms = self.provider.servers.list(search_opts=search_opts)
            for vm in vms:
                print("Deleting VM ({}) : {}".format(vm.name, vm.id))
                vm.delete()

    def delete_floating_ip(self, floating_ip):
        """
        Method deletes a floating IP address
        :param floating_ip:
        :return: None
        """
        return self.provider.floating_ips.delete(floating_ip)

    def start_vm(self, name, group=None, force=None):
        """
        Starts a suspended machine on the target cloud indicated by the id or name
        :param name: Name or ID of the instance to be deleted
        :param group: Security group of the instance to be deleted
        :param force: Force delete option
        :return:
        """

        if self.isUuid(name):
            server = self.provider.servers.get(name)
        else:
            server = self.provider.servers.find(name=name)

        server.start()

    def stop_vm(self, name, group=None, force=None):
        """
        Stops a machine on the target cloud indicated by the id or name
        :param name: Name or ID of the instance to be deleted
        :param group: Security group of the instance to be deleted
        :param force: Force delete option
        :return:
        """

        if self.isUuid(name):
            server = self.provider.servers.get(name)
        else:
            server = self.provider.servers.find(name=name)

        server.stop()

    def get_ips(self, name, group=None, force=None):
        """
        Returns the ip of the instance indicated by name
        :param name:
        :param group:
        :param force:
        :return: IP address of the instance
        """
        if self.isUuid(name):
            server = self.provider.servers.get(name)
        else:
            server = self.provider.servers.find(name=name)

        return self.provider.servers.ips(server)

    def get_floating_ip(self, floating_ip_id):
        """
        Returns the floating ip info associated with floating_ip_id
        :param floating_ip_id:
        :return: floating ip info
        """
        float_ip = self.provider.floating_ips.get(floating_ip_id)
        return float_ip.__dict__["_info"]

    def get_fixed_ip(self, fixed_ip_addr):
        """
        Returns the fixed ip info associated with fixed ip addr
        :param fixed_ip_addr:
        :return: fixed ip info
        """
        fixed_ip = self.provider.fixed_ips.get(fixed_ip_addr)
        return fixed_ip.__dict__["_info"]

    def reserve_fixed_ip(self, fixed_ip_addr):
        """
        Reserves a fixed ip address
        :param fixed_ip_addr:
        :return:
        """
        self.provider.fixed_ips.reserve(fixed_ip_addr)

    def unreserve_fixed_ip(self, fixed_ip_addr):
        """
        Unreserves a fixed ip address
        :param fixed_ip_addr:
        :return:
        """
        self.provider.fixed_ips.unreserve(fixed_ip_addr)

    def create_assign_floating_ip(self, server_name):
        """
        Function creates a new floating ip and associates it with the machine specified.
        :param server_name: Name of the machine to which the floating ip needs to be assigned.
        :return: Floating IP | None if floating ip already assigned.
        """
        ret = None
        allocated_ips = self.list_floating_ips()
        fip = None
        for idx, ipobj in list(allocated_ips.items()):
            if ipobj['instance_id'] is None:
                fip = ipobj['ip']
                break
        if not fip:
            float_pool = self.provider.floating_ip_pools.list()[0].name
            try:
                floating_ip = self.provider.floating_ips.create(pool=float_pool)
                fip = floating_ip.ip
            except:
                ret = "Error: No more floating ips available"
        if fip:
            server = self.provider.servers.find(name=server_name)
            try:
                server.add_floating_ip(fip)
            except Exception as e:
                print(e)
                self.provider.floating_ips.delete(floating_ip)

            ret = fip
        else:
            ret = "Error: failed to create floatingip"
        return ret

    def add_key_to_cloud(self, name, public_key):
        """
        Method to add key to cloud, typically a keypair for openstack.
        :param name: Name of the keypair.
        :param public_key: public key string.
        :return:
        """

        # print("Name=" + name)
        # print("public_key=" + public_key)

        keypair = self.provider.keypairs.create(name, public_key=public_key)
        return keypair

    def delete_key_from_cloud(self, name):
        """
        Method to delete key from cloud, typically a keypair for openstack.
        :param name: Name of the keypair.
        :return:
        """

        # print("Name=" + name)
        # print("public_key=" + public_key)

        keypair = self.provider.keypairs.delete(name)
        return keypair

    def create_floating_ip(self, float_pool):
        """
        Method creates a new floating ip under the specified pool
        :param float_pool:
        :return: Floating IP
        """
        floating_ip = self.provider.floating_ips.create(pool=float_pool)
        return floating_ip.ip

    # TODO: define this
    # noinspection PyProtectedMember,PyProtectedMember
    def get_image(self, **kwargs):

        """
        finds the image based on a query
        TODO: details TBD
        """
        name = kwargs['name']
        if self.isUuid(name):
            return self.provider.images.get(name)._info
        else:
            return self.provider.images.find(name=name)._info

    # TODO: define this
    # noinspection PyProtectedMember,PyProtectedMember
    def get_image_id(self, name):

        """
        finds the image based on a query
        TODO: details TBD
        """
        return self.get_image(name=name)["id"]

    def get_flavor_id(self, name):

        """
        finds the image based on a query
        TODO: details TBD
        """
        return self.get_flavor(id=name)["id"]

    def isUuid(self, name):
        try:
            UUID(name, version=4)
            return True
        except ValueError:
            return False

    # noinspection PyProtectedMember
    def get_flavor(self, **kwargs):
        """
        finds the flavor based on a query
        TODO: details TBD
        """
        if kwargs['id'].isdigit():
            pass
        else:
            kwargs['name'] = kwargs['id']
            del kwargs['id']
        return self.provider.flavors.find(**kwargs)._info

    # noinspection PyProtectedMember,PyProtectedMember
    def get_vm(self, **kwargs):
        """
        finds the flavor based on a query
        TODO: details TBD
        """
        vm_name = kwargs['name']
        if self.isUuid(vm_name):
            return self.provider.servers.get(vm_name)._info
        else:
            return self.provider.servers.find(name=vm_name,
                                              scope="first")._info

    def attributes(self, kind):

        layout = {
            'flavor': {
                'order': [
                    'cm_id',
                    'name',
                    #'user',
                    'ram',
                    'os_flv_disabled',
                    'vcpus',
                    'swap',
                    'os_flavor_acces',
                    'rxtx_factor',
                    'os_flv_ext_data',
                    'disk',
                    'category',
                    'updated_at'
                ],
                'header': [
                    'Id',
                    'Name',
                    #'User',
                    'RAM',
                    'Disabled',
                    'vCPUs',
                    'Swap',
                    'Access',
                    'rxtx_factor',
                    'os_flv_ext_data',
                    'Disk',
                    'Cloud',
                    'updated'
                ]
            },
            'image': {
                'order': [
                    'cm_id',
                    'name',
                    'os_image_size',
                    'metadata__description',
                    'minDisk',
                    'minRam',
                    'status',
                    'category',
                ],
                'header': [
                    'id',
                    'name',
                    'size',
                    'description',
                    'minDisk',
                    'minRam',
                    'status',
                    'cloud',
                ]
            },
            'vm': {
                'order': [
                    'cm_id',
                    'group',
                    'name',
                    'status',
                    'static_ip',
                    'floating_ip',
                    'image',
                    'flavor',
                    'username',
                    'key',
                    'project',
                    'category',
                    'updated_at',
                    'user'
                ],
                'header': [
                    'id',
                    'group',
                    'name',
                    'status',
                    'static_ip',
                    'floating_ip',
                    'username',
                    'image',
                    'flavor',
                    'key',
                    'project',
                    'cloud',
                    'updated_at',
                    'user'
                ]
            },
            'ip': {
                'order': [
                    'name',
                    'static_ip',
                    'floating_ip',
                ],
                'header': [
                    'name',
                    'static_ip',
                    'floating_ip',
                ]
            },
            'floating_ip': {
                'order': [
                    "instance_name",
                    "ip",
                    "pool",
                    "fixed_ip",
                    # "id",
                    # "instance_id",
                    'cloud'
                ],
                'header': [
                    "instance_name",
                    "floating_ip",
                    "floating_ip_pool",
                    "fixed_ip",
                    # "floating_ip_id",
                    # "instance_id",
                    'cloud'
                ],
            },
            'floating_ip_pool': {
                'order': [
                    "name"
                ],
                'header': [
                    "floating_ip_pool"
                ],
            },
            'clouds': {
                'order': [
                    "id",
                    "cloud",
                    "default",
                    "active",
                    "status",
                    "key"
                ],
                'header': [
                    "id",
                    "Cloud",
                    "Default",
                    "Active",
                    "Status",
                    "Key"
                ],
            },
            'limits': {
                'order': [
                    'Name',
                    'Value'
                ],
                'header': [
                    'Name',
                    'Value'
                ]
            },
            'quota': {
                'order': [
                    'Quota',
                    'Limit'
                ],
                'header': [
                    'Quota',
                    'Limit'
                ]
            },
            'secgroup': {
                'order': [
                    'cm_id',
                    'group',
                    'name',
                    'category',
                    'user',
                    'project',
                    'uuid'
                ],
                'header': [
                    'Id',
                    'Group'
                    'Name',
                    'Category',
                    'User',
                    'Tenant',
                    'Uuid'
                ]
            },
            'default': {
                'order': [
                    'user',
                    'category',
                    'name',
                    'value',
                    'created_at',
                    'updated_at'
                ],
                'header': [
                    'user',
                    'category',
                    'name',
                    'value',
                    'created_at',
                    'updated_at'
                ],
            },
            'group': {
                'order': [
                    "name",
                    "member",
                    "user",
                    "category",
                    "type"],
                'header': [
                    "name",
                    "member",
                    "user",
                    "category",
                    "type"]

            },
            'key': {
                'order': [
                    'category',
                    'keypair__name',
                    "type",
                    "comment",
                    "keypair__fingerprint"
                ],
                'header': [
                    "Category",
                    "Name",
                    "Type",
                    "Comment",
                    "Fingerprint"]
            }
        }

        if kind in layout:
            order = layout[kind]["order"]
            header = layout[kind]["header"]
        else:
            order = None
            header = None

        return order, header


# CloudProviderBase.register(CloudProviderOpenstackAPI)

if __name__ == "__main__":
    cloudname = 'kilo'
    d = ConfigDict("cloudmesh.yaml")
    cloud_details = d["cloudmesh"]["clouds"][cloudname]

    cp = CloudProviderOpenstackAPI(cloudname, cloud_details)

    d = {'name': '390792c3-66a0-4c83-a0d7-c81e1c787710'}
    pprint(cp.list_quota(cloudname))

    pprint(cp.list_key(cloudname))



"""Ascii menu class"""
from __future__ import print_function
from cloudmesh_client.common.Printer import Printer

from builtins import input


def ascii_menu(title=None, menu_list=None):
    """
    creates a simple ASCII menu from a list of tuples containing a label
    and a functions reference. The function should not use parameters.
    :param title: the title of the menu
    :param menu_list: an array of tuples [('label', f1), ...]
    """
    if not title:
        title = "Menu"

    n = len(menu_list)

    def display():
        index = 1
        print()
        print(title)
        print(len(title) * "=")
        print()
        for label, function in menu_list:
            print("    {0:>3} - {1}".format(index, label))
            index += 1
        print("    q - quit")
        print()
        print()

    display()
    running = True
    while running:
        result = input("Select between {0} - {1}: ".format(1, n))
        print("<{0}>".format(result))
        if result.strip() in ["q"]:
            running = False
        else:
            try:
                result = int(result) - 1
                if 0 <= result < n:
                    (label, f) = menu_list[result]
                    print("EXECUTING:", label, f.__name__)
                    f()
                else:
                    print("ERROR: wrong selection")
            except Exception as e:
                print("ERROR: ", e)
        display()


def menu_return_num(title=None, menu_list=None, tries=1, with_display=True):
    """
    creates a simple ASCII menu from a list of labels
    :param title: the title of the menu
    :param menu_list: a list of labels to choose
    :param tries: num of tries till discard
    :return: choice num (head: 0), quit: return 'q'
    """
    if not title:
        title = "Menu"

    n = len(menu_list)

    def display():
        index = 1
        print()
        print(title)
        print(len(title) * "=")
        print()
        for label in menu_list:
            print("    {0:>3} - {1}".format(index, label))
            index += 1
        print("      q - quit")
        print()
        print()

    display()
    while tries > 0:
        # display()
        result = input("Select between {0} - {1}: ".format(1, n))
        if result == "q":
            return 'q'
        else:
            try:
                result = int(result)
            except:
                print("invalid input...")
                tries -= 1
                continue
            if 0 < result <= n:
                print("choice {0} selected.".format(result))
                return result - 1
            else:
                print("ERROR: wrong selection")

    return 'q'


def num_choice(n, tries=1):

    while tries > 0:
        # display()
        result = input("Select between {0} - {1}: ".format(1, n))
        if result == "q":
            return 'q'
        else:
            try:
                result = int(result)
            except:
                print("invalid input...")
                tries -= 1
                continue
            if 0 < result <= n:
                print("choice {0} selected.".format(result))
                return result - 1
            else:
                print("ERROR: wrong selection")

    return 'q'


def dict_choice(d):
    if d is None:
        return None

    elements = dict(d)
    i = 1
    for e in d:
        elements[e]["id"] = i
        i += 1
    #   pprint(d)
    if elements != {}:
        # noinspection PyPep8
        print(Printer.write(elements,
                            order=["id",
                                  "name",
                                  "comment",
                                  "uri",
                                  "fingerprint",
                                  "source"],
                            output="table",
                            sort_keys=True))
    else:
        print("ERROR: No keys in the database")
        return

    n = num_choice(i - 1, tries=10) + 1
    element = None
    for e in elements:
        if str(elements[e]["id"]) is str(n):
            element = elements[e]
            break
    return element



from __future__ import print_function

import socket
from uuid import UUID

from cloudmesh_client.shell.console import Console
from cloudmesh_client.common.Printer import Printer
from cloudmesh_client.cloud.ListResource import ListResource
from cloudmesh_client.common.Printer import Printer
from cloudmesh_client.cloud.iaas.CloudProvider import CloudProvider
from cloudmesh_client.db.CloudmeshDatabase import CloudmeshDatabase

from pprint import pprint
from builtins import input


# noinspection PyBroadException,PyPep8Naming
class Network(ListResource):
    cm = CloudmeshDatabase()

    @classmethod
    def get_fixed_ip(cls, cloudname, fixed_ip_addr, output='table'):
        """
        Method retrieves fixed ip info
        :param cloudname:
        :param fixed_ip_addr:
        :return: fixed_ip_info
        """
        try:
            cloud_provider = CloudProvider(cloudname).provider
            result = cloud_provider.get_fixed_ip(fixed_ip_addr=fixed_ip_addr)

            return Printer.attribute(result,
                                     header=[
                                         "name",
                                         "value"
                                     ],
                                     output=output)
        except Exception as ex:
            Console.error(ex.message)

        return

    @classmethod
    def get_floating_ip(cls, cloudname, floating_ip_or_id, output='table'):
        """
        Method to get floating ip info
        :param cloudname:
        :param floating_ip_or_id:
        :return: floating ip info
        """
        try:
            cloud_provider = CloudProvider(cloudname).provider
            result = None

            # check if argument is ip or uuid
            if cls.isIPAddr(ip_or_id=floating_ip_or_id):
                # get floating ip list
                floating_ips = cls.get_floating_ip_list(cloudname)
                for floating_ip in list(floating_ips):
                    ip_addr = floating_ip["ip"]

                    # if argument ip matches floating ip addr
                    if ip_addr == floating_ip_or_id:
                        result = floating_ip
                        break
            else:
                # find by floating ip uuid
                result = cloud_provider.get_floating_ip(floating_ip_id=floating_ip_or_id)

            # Could not find floating IP from given args
            if result is None:
                return None

            instance_id = result["instance_id"]
            instance_name = None

            if instance_id is not None:
                # lookup instance_name from id
                instance_name = cls.find_instance_name(cloudname=cloudname,
                                                       instance_id=instance_id)

            # add instance_name to dict
            result["instance_name"] = instance_name

            # add misc details to response
            result["cloud"] = cloudname
            result["user"] = cloud_provider.cloud_details["credentials"]["OS_USERNAME"]
            result["project"] = cloud_provider.cloud_details["credentials"]["OS_TENANT_NAME"]

            return Printer.attribute(result,
                                     header=[
                                         "name",
                                         "value"
                                     ],
                                     output=output)
        except Exception:
            # auto detect floating-ip-id
            floating_ips = cls.get_floating_ip_list(cloudname)
            # for each floating-ip from list
            for floating_ip in list(floating_ips):
                if floating_ip["id"].startswith(floating_ip_or_id) or \
                        floating_ip["ip"].startswith(floating_ip_or_id):
                    # confirm choice with user
                    print("Did you mean floating-ip [{}] ? (y/n)".format(floating_ip["ip"]))
                    choice = input().lower()
                    # if yes, return dict
                    if choice == 'y':
                        return Printer.attribute(floating_ip,
                                                 header=[
                                                     "name",
                                                     "value"
                                                 ],
                                                 output=output)
                        # Console.error(ex.message)
        return

    @classmethod
    def reserve_fixed_ip(cls, cloudname, fixed_ip_addr):
        """
        Reserve a fixed ip address
        :param cloudname:
        :param fixed_ip_addr:
        :return:
        """
        try:
            cloud_provider = CloudProvider(cloudname).provider
            cloud_provider.reserve_fixed_ip(fixed_ip_addr=fixed_ip_addr)
            return "Success."
        except Exception as ex:
            Console.error(ex.message)


        return

    @classmethod
    def unreserve_fixed_ip(cls, cloudname, fixed_ip_addr):
        """
        Unreserve a fixed ip address
        :param cloudname:
        :param fixed_ip_addr:
        :return:
        """
        try:
            cloud_provider = CloudProvider(cloudname).provider
            cloud_provider.unreserve_fixed_ip(fixed_ip_addr=fixed_ip_addr)
            return "Success."
        except Exception as ex:
            Console.error(ex.message)
            return ex

    @classmethod
    def associate_floating_ip(cls, cloudname, instance_name, floating_ip):
        """
        Method to associate floating ip to an instance
        :param cloudname:
        :param instance_name:
        :param floating_ip:
        :return:
        """
        try:
            cloud_provider = CloudProvider(cloudname).provider
            # Find the server instance
            server = cloud_provider.provider.servers.find(name=instance_name)
            # Add the floating ip to the instance
            server.add_floating_ip(floating_ip)
            return "Success."
        except Exception as ex:
            if "already has a floating" in ex.message:
                Console.error("VM has already floating ip", traceflag=False)
            else:
                Console.error(ex.message)
                return ex
        pass

    @classmethod
    def disassociate_floating_ip(cls, cloudname, instance_name, floating_ip):
        """
        Disassociates a floating ip from an instance
        :param cloudname:
        :param instance_name:
        :param floating_ip:
        :return:
        """
        try:
            cloud_provider = CloudProvider(cloudname).provider
            # Find the server instance
            server = cloud_provider.provider.servers.find(name=instance_name)
            # Remove the floating ip from the instance
            server.remove_floating_ip(floating_ip)
            # Release the floating ip to the pool
            cls.delete_floating_ip(cloudname=cloudname,
                                   floating_ip_or_id=floating_ip)
            return "Success."
        except Exception as ex:
            Console.error(ex.message)
            return ex
        pass

    @classmethod
    def create_assign_floating_ip(cls, cloudname, instance_name):
        """
        Method to create a new floating-ip
        and associate it with the instance
        :param cloudname: cloud
        :param instance_name: name of instance
        :return: floating_ip
        """
        try:
            cloud_provider = CloudProvider(cloudname).provider
            floating_ip = cloud_provider.create_assign_floating_ip(instance_name)
            return floating_ip
        except Exception as ex:
            Console.error(ex.message)
            return

    @classmethod
    def create_floating_ip(cls, cloudname, floating_pool=None):
        """
        Method to create a floating ip address under a pool
        :param cloudname:
        :param floating_pool:
        :return: floating ip addr
        """
        try:
            cloud_provider = CloudProvider(cloudname).provider
            # If floating pool is not given,
            # get first from list
            if floating_pool is None:
                floating_pool = cloud_provider.provider.floating_ip_pools.list()[0].name
                Console.ok("Floating pool not provided, selected [{}] as the pool."
                           .format(floating_pool))

            floating_ip = cloud_provider.create_floating_ip(float_pool=floating_pool)
            return floating_ip
        except Exception as ex:
            Console.error(ex.message)
            return

    @classmethod
    def delete_floating_ip(cls, cloudname, floating_ip_or_id):
        """
        Method to delete a floating ip address
        :param cloudname:
        :param floating_ip_or_id:
        :return:
        """
        try:
            cloud_provider = CloudProvider(cloudname).provider
            floating_ip_dict = None

            # check if argument is ip or uuid
            if cls.isIPAddr(ip_or_id=floating_ip_or_id):
                # get floating ip list
                floating_ips = cls.get_floating_ip_list(cloudname)
                for floating_ip in list(floating_ips):
                    ip_addr = floating_ip["ip"]

                    # if argument ip matches floating ip addr
                    if ip_addr == floating_ip_or_id:
                        floating_ip_dict = floating_ip
                        break
            else:
                # find by floating ip uuid
                floating_ip_dict = cloud_provider.get_floating_ip(floating_ip_id=floating_ip_or_id)

            # Could not find floating IP from given args
            if floating_ip_dict is None:
                return None

            # Delete the floating ip; returns None if success
            result = cloud_provider.delete_floating_ip(floating_ip_dict["id"])

            if not result:
                return "Floating IP [{}] deleted successfully!" \
                    .format(floating_ip_dict["ip"])

        except Exception as ex:
            Console.error(ex.message)
            return

    @classmethod
    def list_floating_ip(cls, cloudname, output='table'):
        """
        Method to list floating ips
        :param cloudname:
        :return: floating ip list
        """
        try:
            floating_ips = cls.get_floating_ip_list(cloudname)

            for floating_ip in list(floating_ips.values()):
                # Get instance_id associated with instance
                instance_id = floating_ip["instance_id"]

                if instance_id is not None:
                    try:
                        instance_name = cls.find_instance_name(cloudname=cloudname,
                                                               instance_id=instance_id)
                        # Assign it to the dict
                        floating_ip["instance_name"] = instance_name
                        floating_ip["cloud"] = cloudname
                    except Exception as ex:
                        Console.error(ex.message)
                        continue
                else:
                    # If no instance associated, keep None
                    floating_ip["instance_name"] = None

            (order, header) = CloudProvider(cloudname).get_attributes("floating_ip")

            return Printer.write(floating_ips,
                                 order=order,
                                 header=header,
                                 output=output)
        except Exception as ex:
            Console.error(ex.message)

        return

    @classmethod
    def list_unused_floating_ip(cls, cloudname, output='table'):
        """
        Method to list unused floating ips
        These floating ips are not associated with any instance
        :param cloudname:
        :return: floating ip list
        """
        try:
            # fetch unused floating ips
            floating_ips = cls.get_unused_floating_ip_list(cloudname)

            # print the output
            return Printer.write(floating_ips,
                                 order=[
                                     "ip",
                                     "pool",
                                     "id",
                                     "cloud"
                                 ],
                                 header=[
                                     "floating_ip",
                                     "floating_ip_pool",
                                     "floating_ip_id",
                                     "cloud"
                                 ],
                                 output=output)
        except Exception as ex:
            Console.error(ex.message)

        return

    @classmethod
    def list_floating_ip_pool(cls, cloudname):
        """
        Method to list floating ip pool
        :param cloudname:
        :return:
        """
        try:
            cloud_provider = CloudProvider(cloudname).provider
            floating_ip_pools = cloud_provider.list_floating_ip_pools()

            (order, header) = CloudProvider(cloudname).get_attributes("floating_ip_pool")

            return Printer.write(floating_ip_pools,
                                 order=order,
                                 header=header)

        except Exception as ex:
            Console.error(ex.message)
        pass

    @classmethod
    def isIPAddr(cls, ip_or_id):
        """
        Method to check if argument is IP address or notS
        :param ip_or_id:
        :return:
        """
        try:
            socket.inet_aton(ip_or_id)
            return True
        except:
            return False

    @classmethod
    def get_unused_floating_ip_list(cls, cloudname):
        """
        Method to get the unused floating IP list
        :param cloudname:
        :return: floating_ips
        """
        try:
            cloud_provider = CloudProvider(cloudname).provider
            floating_ips = cloud_provider.list_floating_ips()
            unused_floating_ips = list()

            for floating_ip in list(floating_ips.values()):
                if floating_ip["fixed_ip"] is None and \
                                floating_ip["instance_id"] is None:
                    # add to unused list
                    floating_ip["cloud"] = cloudname
                    unused_floating_ips.append(floating_ip)

            return unused_floating_ips
        except Exception as ex:
            Console.error(ex.message)

    @classmethod
    def get_floating_ip_list(cls, cloudname):
        """
        Method to get the floating IP list
        :param cloudname:
        :return: floating_ips
        """
        try:
            cloud_provider = CloudProvider(cloudname).provider
            floating_ips = cloud_provider.list_floating_ips()
            return floating_ips
        except Exception as ex:
            Console.error(ex.message)

    @classmethod
    def find_instance_name(cls, **kwargs):
        """
        Method to find instance name
        :param kwargs:
        :return: instance_name
        """
        cloudname = kwargs["cloudname"]
        instance_id = kwargs["instance_id"]

        # Cloudmesh database instance

        # Lookup instance details from db
        instance_dict = cls.cm.find(kind="vm", category=cloudname,
                                    uuid=instance_id)

        # Get instance_name for vm
        if len(instance_dict) > 0:
            instance_name = list(instance_dict)[0]["name"]
            return instance_name

    @classmethod
    def get_instance_dict(cls, **kwargs):
        """
        Method to get instance dict
        :param kwargs:
        :return: instance dict
        """
        cloudname = kwargs["cloudname"]
        instance_id = kwargs["instance_id"]

        # Cloudmesh database instance

        # Lookup instance details from db
        if cls.isUuid(instance_id):
            instance_dict = cls.cm.find(kind="vm", category=cloudname,
                                        uuid=instance_id)
        else:
            instance_dict = cls.cm.find(kind="vm", category=cloudname,
                                        name=instance_id)

        # Instance not found in DB
        if cls.isDictEmpty(instance_dict):
            # auto detect instance_id feature
            vms = cls.cm.find(kind="vm", category=cloudname)
            # check for each instance in db
            for vm in list(vms):
                # if match found in either name/id
                if vm["uuid"].startswith(instance_id) or \
                        vm["name"].startswith(instance_id):
                    # confirm choice with user
                    print("Did you mean instance [{}] ? (y/n)".format(vm["name"]))
                    choice = input().lower()
                    # if yes, return dict
                    if choice == 'y':
                        return vm
            return None
        else:
            return list(instance_dict)[0]

    @classmethod
    def find_assign_floating_ip(cls, cloudname, instance_id):
        instance_dict = cls.get_instance_dict(cloudname=cloudname,
                                              instance_id=instance_id)
        # Instance not found
        if instance_dict is None:
            Console.error("Instance [{}] not found in the cloudmesh database!"
                          .format(instance_id))
            return None

        instance_name = instance_dict["name"]

        # Get an unused ip from pool if exist and associate
        unused_floating_ips = cls.get_unused_floating_ip_list(cloudname=cloudname)

        if unused_floating_ips:
            floating_ip = unused_floating_ips[0]["ip"]
            result = cls.assign_floating_ip(cloudname=cloudname,
                                            instance_id=instance_id,
                                            floating_ip=floating_ip)
            if result is None:
                Console.error("IP {} could not be assigned to VM {} "
                          .format( floating_ip, instance_name))
                return None
        else:
            # create a new ip and associate
            floating_ip = cls.create_assign_floating_ip(cloudname=cloudname,
                                                            instance_name=instance_name)

        return floating_ip

    @classmethod
    def assign_floating_ip(cls, cloudname, instance_id, floating_ip):
        # find instance in db
        instance_dict = cls.get_instance_dict(cloudname=cloudname,
                                              instance_id=instance_id)
        # Instance not found
        if instance_dict is None:
            Console.error("Instance [{}] not found in the cloudmesh database!"
                          .format(instance_id))
            return None

        instance_name = instance_dict["name"]
        result = cls.associate_floating_ip(cloudname=cloudname,
                                           instance_name=instance_name,
                                           floating_ip=floating_ip)
        return result

    @classmethod
    def isUuid(cls, argument):
        """
        Method to check if arg is an UUID
        :param argument:
        :return:
        """
        try:
            UUID(argument, version=4)
            return True
        except ValueError:
            return False

    @classmethod
    def isDictEmpty(cls, dictionary):
        """
        Method to test empty Dict
        :param dictionary:
        :return:
        """
        if bool(dictionary):
            return False
        else:
            return True




from StringIO import StringIO
from collections import defaultdict



__all__ = ['Inventory', 'Node']


class Node(object):

    def __init__(self, name, address=None, user=None):
        self._name = name
        self._address = address or name
        self._user = user or None
        self._variables = dict()

    ################################################################ main api

    def add_var(self, key, value):
        """Add a variable and value to a node

        :param str key:
        :param str value:
        """

        self._variables[key] = value

    ################################################################ properties

    @property
    def name(self):
        return self._name

    @property
    def address(self):
        return self._address

    @property
    def user(self):
        return self._user

    @property
    def variables(self):
        """Get the dictionary of variables

        :rtype: :class:`dict`
        """

        d = dict(
            ansible_ssh_host=self._address
        )
        d.update(self._variables)
        if self.user:
            d['ansible_ssh_user'] = self.user

        return d


################################################################################


class Inventory(object):
    """Build an inventory by dynamically adding :class:`Host`s to :class:`Group`s"""

    def __init__(self):
        self._groups = defaultdict(list)
        self._nodes = set(['all'])

    @classmethod
    def from_cluster(cls, cluster):
        """Creates an inventory from a :class:`Cluster`

        :param cls:
        :param cluster:
        :returns:
        :rtype:
        """

        inventory = cls()
        for instance in cluster:
            node = Node(
                name = instance.name,
                address = instance.floating_ip or instance.static_ip,
                user = instance.username or None
            )
            inventory.add_node(node)

        return inventory

    def add_node(self, node, *groupnames):
        """Add a host to the inventory

        :param str groupname:
        :param Node host:
        """

        groupnames = groupnames or ['all']

        for groupname in groupnames:
            self._groups[groupname].append(node)

        self._nodes.add(node)

    def ini(self):
        """Generates the ansible inventory file

        :returns: the inventory as ini format
        :rtype: :class:`str`
        """

        builder = StringIO()
        for groupname, nodes in self._groups.iteritems():
            builder.write('[{}]\n'.format(groupname))
            for node in nodes:
                builder.write('{}'.format(node.name))
                for k, v in node.variables.iteritems():
                    builder.write(' {}="{}"'.format(k, v))
                builder.write('\n')
            builder.write('\n')

        ini = builder.getvalue().strip()
        builder.close()

        return ini

        # ################################
        # ## initialize the root Group, ## needed for the call to InventoryParser constructor

        # groups = {'all': Group('all')}

        # ################################
        # ## write the inventory to a temp file, then read it in using ansible API

        # with NamedTemporaryFile() as fd:
        #     fd.write(ini)
        #     fd.flush()
        #     inventory = InventoryParser(None, groups, filename=fd.name)
        #     return inventory




if __name__ == '__main__':
    from cloudmesh_client.default import Default
    cluster = Default.active_cluster
    i = Inventory.from_cluster(cluster)
    print(i.ini())

    n1 = Node('foo', address='129.114.110.195', user='cc')
    n2 = Node('bar', address='129.114.110.127', user='cc')
    n3 = Node('baz', address='129.114.111.200', user='cc')

    inventory = Inventory()
    inventory.add_node(n1, 'a', 'b')
    inventory.add_node(n2, 'c')
    inventory.add_node(n3, 'b', 'c')

    i = inventory.ini()
    print(i)
